"""Tests of the caching module."""

from pathlib import Path

import attrs
import numpy as np
import pytest

from py21cmfast import (
    Coeval,
    InitialConditions,
    InputParameters,
    IonizedBox,
    PerturbedField,
    compute_initial_conditions,
)
from py21cmfast.io import caching, h5
from py21cmfast.wrapper import outputs


def create_full_run_cache(cachedir: Path) -> caching.RunCache:
    inputs = InputParameters.from_template(
        "latest",
        random_seed=12345,
        node_redshifts=np.arange(12, 38, 3.0)[::-1],
    ).evolve_input_structs(HII_DIM=10, DIM=20, BOX_LEN=75.0, ZPRIME_STEP_FACTOR=1.3)
    cache = caching.RunCache.from_inputs(inputs, caching.OutputCache(cachedir))

    for fldname, fld in attrs.asdict(cache, recurse=False).items():
        if isinstance(fld, dict):
            for z, fname in fld.items():
                o = getattr(outputs, fldname).new(redshift=z, inputs=inputs)
                o._init_arrays()

                # Go through each array and set it to be "computed" so we can trick
                # the writer into writing it out to file.
                for k, v in o.arrays.items():
                    setattr(o, k, v.with_value(v.value))

                # Mock the primitive fields as well...
                for fld in o.struct.primitive_fields:
                    setattr(o, fld, 0.0)

                h5.write_output_to_hdf5(o, fname)
        elif fldname == "InitialConditions":
            o = outputs.InitialConditions.new(inputs=inputs)
            o._init_arrays()
            for k, v in o.arrays.items():
                setattr(o, k, v.with_value(v.value))
            h5.write_output_to_hdf5(o, fld)
    return cache


@pytest.fixture(scope="module")
def full_run_cache(tmp_path_factory):
    return create_full_run_cache(tmp_path_factory.mktemp("full_run_cache"))


@pytest.fixture(scope="module")
def partial_run_cache(tmp_path_factory):
    cache = create_full_run_cache(tmp_path_factory.mktemp("partial_run_cache"))
    cache.PerturbedField[cache.inputs.node_redshifts[-1]].unlink()
    return cache


class TestRunCache:
    """Tests of the RunCache class."""

    def test_optional_boxes(self, tmp_path: Path):
        """Test that the RunCache can be created with optional boxes."""
        inputs = InputParameters.from_template("latest-dhalos", random_seed=12345)
        cache = caching.RunCache.from_inputs(inputs, caching.OutputCache(tmp_path))
        print(attrs.asdict(cache).keys(), flush=True)

        assert isinstance(cache.HaloBox, dict)
        assert isinstance(cache.HaloCatalog, dict)
        assert isinstance(cache.InitialConditions, Path)
        assert isinstance(cache.PerturbedField, dict)
        assert isinstance(cache.IonizedBox, dict)
        assert isinstance(cache.BrightnessTemp, dict)
        assert isinstance(cache.TsBox, dict)
        assert isinstance(cache.XraySourceBox, dict)

        assert len(cache.HaloBox) == len(inputs.node_redshifts)

        inputs = InputParameters.from_template("simple", random_seed=12345)
        cache = caching.RunCache.from_inputs(inputs, caching.OutputCache(tmp_path))

        assert cache.HaloBox is None

    def test_from_example_file(self, full_run_cache: caching.RunCache):
        """Test the from_example_file classmethod."""
        for fld in attrs.asdict(full_run_cache, recurse=False).values():
            if isinstance(fld, dict):
                for fname in fld.values():
                    cache2 = caching.RunCache.from_example_file(fname)
                    assert full_run_cache == cache2

    def test_bad_example_file(self, perturbed_field: PerturbedField, tmpdirec):
        """Test that the RunCache fails to construct when given a non-cache file."""
        badpath = tmpdirec / Path("testpf.h5")
        h5.write_output_to_hdf5(perturbed_field, badpath)

        with pytest.raises(
            ValueError, match="does not seem to be within a cache structure"
        ):
            caching.RunCache.from_example_file(badpath)

    def test_is_complete_at(self, full_run_cache, partial_run_cache):
        """Test that is_complete_at works as expected."""
        assert full_run_cache.is_complete_at(z=full_run_cache.inputs.node_redshifts[-1])
        assert full_run_cache.is_complete_at(
            index=len(full_run_cache.inputs.node_redshifts) - 1
        )

        assert not partial_run_cache.is_complete_at(
            z=partial_run_cache.inputs.node_redshifts[-1]
        )
        assert partial_run_cache.is_complete_at(index=0)

    def test_get_output_struct_at_z(self, full_run_cache):
        """Test that get_output_struct_at_z works as expected."""
        cache = full_run_cache
        for name, filedict in attrs.asdict(cache, recurse=False).items():
            if not isinstance(filedict, dict):
                continue

            for idx, z in enumerate(cache.inputs.node_redshifts):
                output = cache.get_output_struct_at_z(kind=name, z=z + 1e-3)
                assert isinstance(output, getattr(outputs, name))
                assert output.redshift == z

                output = cache.get_output_struct_at_z(kind=name, z=z)
                assert isinstance(output, getattr(outputs, name))
                assert output.redshift == z

                output = cache.get_output_struct_at_z(kind=name, index=idx)
                assert isinstance(output, getattr(outputs, name))
                assert output.redshift == z

                output = cache.get_output_struct_at_z(
                    kind=getattr(outputs, name), index=idx
                )
                assert isinstance(output, getattr(outputs, name))
                assert output.redshift == z

    def test_get_output_struct_at_z_raises(self, full_run_cache):
        """Test that get_output_struct_at_z raises as expected."""
        cache = full_run_cache

        with pytest.raises(ValueError, match="Cannot specify both z and index"):
            cache.get_output_struct_at_z(kind="PerturbedField", z=3.0, index=4)

        with pytest.raises(ValueError, match="Unknown output kind"):
            cache.get_output_struct_at_z(kind="UnknownKind", z=3.0)

        with pytest.raises(ValueError, match="No output struct found"):
            cache.get_output_struct_at_z(kind="TsBox", z=0.0)

    def test_get_all_boxes_at_z(self, full_run_cache):
        """Test get_all_boxes_at_z functionality."""
        cache = full_run_cache

        for z in cache.inputs.node_redshifts:
            boxes = cache.get_all_boxes_at_z(z)
            assert len(boxes) == 4  # number of structs with redshifts (PF, Ts, IB, BT)
            for b in boxes.values():
                assert b.redshift == z

    def test_get_coeval_at_z(self, full_run_cache):
        """Test get_coeval_at_z functionality."""
        cache = full_run_cache

        for z in cache.inputs.node_redshifts:
            coeval = cache.get_coeval_at_z(z)
            assert isinstance(coeval, Coeval)
            assert coeval.redshift == z

    def test_is_complete(self, full_run_cache, partial_run_cache):
        """Test that complete boxes are complete and incomplete boxes are not."""
        assert full_run_cache.is_complete()

        assert not partial_run_cache.is_complete()


class TestOutputCache:
    """Tests of the OutputCache class."""

    def test_readability(
        self,
        ic: InitialConditions,
        cache: caching.OutputCache,
        default_input_struct: InputParameters,
        ionize_box: IonizedBox,
        perturbed_field: PerturbedField,
    ):
        """Test that each type of field can be read from cache."""
        for field in (ic, perturbed_field, ionize_box):
            kwargs = {"inputs": default_input_struct}
            if field.__class__ != InitialConditions:
                kwargs["redshift"] = field.redshift
            field2 = field.__class__.new(**kwargs)
            existing = cache.find_existing(field2)

            assert existing is not None
            assert existing.exists()

            field2 = cache.load(field2)

            assert field is not field2
            assert field == field2

    def test_read_from_filename(
        self, default_input_struct, perturbed_field, ionize_box, ic, cache
    ):
        """Test that each type of field can be read using a filename."""
        for field in (ic, perturbed_field, ionize_box):
            kwargs = {
                "inputs": default_input_struct,
                "kind": field.__class__.__name__,
            }
            if field.__class__ != InitialConditions:
                kwargs["redshift"] = field.redshift
            dlist = cache.list_datasets(**kwargs)

            assert len(dlist) == 1
            field2 = h5.read_output_struct(dlist[0])
            assert field == field2
            assert field is not field2

    def test_read_wildcards(self, perturbed_field, ionize_box, ic, cache):
        """Test that we can read files without specific inputs."""
        dlist = cache.list_datasets()

        assert len(dlist) >= 3
        for field in (ic, perturbed_field, ionize_box):
            filename = cache.get_path(field)
            assert filename in dlist

    def test_match_seed(
        self, cache: caching.OutputCache, default_input_struct: InputParameters
    ):
        """Test that changing the random seed results in a different cache location."""
        ic2 = InitialConditions.new(inputs=default_input_struct.clone(random_seed=3))

        # This fails because we've set the seed and it's different to the existing one.
        with pytest.raises(IOError, match="No cache exists for"):
            cache.load(ic2)

    def test_astro_param_change(
        self, default_input_struct, ic, perturbed_field, ionize_box, cache
    ):
        """Test that changing parameters affects IonizedBox but not ICs or PerturbedField."""
        input_change = default_input_struct.evolve_input_structs(
            F_ESC10=-1.5,
        )

        ic2 = InitialConditions.new(inputs=input_change)
        pf2 = PerturbedField.new(inputs=input_change, redshift=perturbed_field.redshift)
        ib2 = IonizedBox.new(inputs=input_change, redshift=ionize_box.redshift)

        for f2, f in zip((ic2, pf2), (ic, perturbed_field), strict=False):
            existing = cache.find_existing(f2)

            assert existing is not None
            assert existing.exists()

            f2 = cache.load(f2)

            assert f == f2
            assert f is not f2

        assert not cache.find_existing(ib2)

    def test_regeneration(
        self,
        default_input_struct,
        ic,
        cache,
        mocker,
    ):
        """Test that the rengerate keyword works as intended, skipping calculation and loading from cache."""
        reader_spy = mocker.spy(h5, "read_output_struct")
        compute_spy = mocker.spy(InitialConditions, "compute")

        ics_noregen = compute_initial_conditions(
            inputs=default_input_struct, regenerate=False, cache=cache
        )
        assert reader_spy.call_count == 1
        assert compute_spy.call_count == 0
        assert ics_noregen == ic

        reader_spy.reset_mock()
        compute_spy.reset_mock()

        ics_regen = compute_initial_conditions(
            inputs=default_input_struct, regenerate=True, cache=cache
        )
        assert reader_spy.call_count == 0
        assert compute_spy.call_count == 1
        assert ics_regen == ic


@pytest.mark.parametrize(
    "inp_type",
    [
        "random_seed",
        "node_redshifts",
        "simulation_options",
        "matter_options",
        "cosmo_params",
        "astro_params",
        "astro_options",
    ],
)
def test_hash_for_different_inputs(default_input_struct, inp_type):
    """Test that the hash table is different for different inputs."""
    new_kwargs = {
        "random_seed": 1,
        "node_redshifts": (30.0, 35.0),
        "simulation_options": {"BOX_LEN": 300},
        "matter_options": {"SOURCE_MODEL": "L-INTEGRAL"},
        "cosmo_params": {"hlittle": 0.7},
        "astro_params": {"L_X": 38.0},
        "astro_options": {"USE_CMB_HEATING": False},
    }
    hash_default = caching.OutputCache()._get_hashes(default_input_struct)
    if inp_type not in ["random_seed", "node_redshifts"]:
        new_input = getattr(default_input_struct, inp_type).clone(
            **new_kwargs[inp_type]
        )
    else:
        new_input = new_kwargs[inp_type]
    new_inputs = default_input_struct.clone(**{inp_type: new_input})
    new_hash = caching.OutputCache()._get_hashes(new_inputs)

    assert new_hash != hash_default



"""Module that contains the command line app."""

import logging
import uuid
import warnings
from dataclasses import dataclass, field, fields
from pathlib import Path
from typing import Annotated, Literal

import attrs
import matplotlib.pyplot as plt
import numpy as np
from cyclopts import App, Group, Parameter
from cyclopts import types as cyctp
from cyclopts import validators as vld
from rich import box
from rich.columns import Columns
from rich.console import Console, group
from rich.logging import RichHandler
from rich.panel import Panel
from rich.rule import Rule
from rich.table import Table
from rich.text import Text

from . import __version__, plotting
from ._templates import TOMLMode, list_templates, write_template
from .drivers import coeval as cvlmodule
from .drivers.coeval import generate_coeval
from .drivers.lightcone import run_lightcone
from .drivers.single_field import compute_initial_conditions
from .input_serialization import convert_inputs_to_dict
from .io.caching import CacheConfig, OutputCache, RunCache
from .lightconers import RectilinearLightconer
from .wrapper.inputs import (
    AstroOptions,
    AstroParams,
    CosmoParams,
    InputParameters,
    InputStruct,
    MatterOptions,
    SimulationOptions,
)

cns = Console()

FORMAT = "%(message)s"
logging.basicConfig(
    level="WARNING",
    format=FORMAT,
    datefmt="[%X]",
    handlers=[RichHandler(rich_tracebacks=True, console=cns)],
)
logger = logging.getLogger("py21cmfast")

AVAILABLE_TEMPLATES: list[str] = [
    str(alias) for tmpl in list_templates() for alias in tmpl["aliases"]
] + ["defaults"]

app = App()
app.command(
    cfg := App(name="template", help="Manage 21cmFAST configuration files/templates.")
)
app.command(run := App(name="run", help="Run 21cmFAST simulations."))
app.command(dev := App(name="dev", help="Run development tasks."))
app.command(pred := App(name="predict", help="Predict properties of simulations"))


def print_banner():
    """Print a cool banner for 21cmFAST using rich."""
    panel = Panel(
        Text.from_markup("[orange]:duck: 21cmFAST :duck:", justify="center"),
        subtitle=f"v{__version__}",
        style="cyan",
        padding=1,
        highlight=True,
        box=box.DOUBLE_EDGE,
    )
    cns.print(panel)
    cns.print()


_paramgroup = Group(
    "Simulation Parameters (set one)",
    validator=(vld.MutuallyExclusive(),),
)


@Parameter(name="*")
@dataclass(frozen=True, kw_only=True)
class ParameterSelection:
    """Common options for choosing simulation parameters.

    You can either specify a TOML file, or a template name.
    """

    param_file: Annotated[
        list[cyctp.ExistingTomlPath], Parameter(group=_paramgroup)
    ] = None
    "Path to a TOML configuration file (can be generated with `21cmfast template create`)."

    template: Annotated[
        list[Literal[*AVAILABLE_TEMPLATES]],
        Parameter(group=_paramgroup, alias=("--base-template"), consume_multiple=True),
    ] = field(default_factory=lambda: ["defaults"])

    "The name of a valid builtin template (see available with `21cmfast template avail`)."


@Parameter(name="*")
@dataclass(kw_only=True)
class RunParams:
    """Common parameters for run functions."""

    param_selection: ParameterSelection = ParameterSelection()

    seed: int = 42
    "Random seed used to generate data."

    regenerate: Annotated[bool, Parameter(name=["--regenerate", "--regen"])] = False
    "Whether to regenerate all data, even if found in cache."

    cachedir: cyctp.ExistingDirectory = Path()
    (
        "Where to write and search for cached items and output fullspec configuration "
        "files. Note that caches will be in hash-style folders inside this folder."
    )
    cache_strategy: Literal["on", "off", "noloop", "dmfield", "last_step_only"] = (
        "dmfield"
    )
    """A strategy for which fields to cache (only used for coeval and lightcones).
Options are: (on) cache everything, (off) cache nothing (noloop) cache only boxes
outside the astrophysics evolution loop (dmfield) alias for noloop (last_step_only)
cache only boxes that are required more than one step away
    """

    outcfg: Annotated[
        Path,
        Parameter(validator=(vld.Path(file_okay=False, dir_okay=False, ext=("toml",)))),
    ] = None
    "A filepath where the full configuration TOML of the run can be written. Random hash if not specified."

    verbosity: Annotated[
        Literal["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"],
        Parameter(alias=("-v", "--v")),
    ] = "WARNING"
    "How much information to print out while running the simulation."

    progress: bool = True
    "Whether to display a progress bar as the simulation runs."


def _param_cls_factory(cls: type[InputStruct]) -> type:
    out = attrs.make_class(
        cls.__name__,
        {
            fld.alias: attrs.field(default=None, type=fld.type)
            for fld in attrs.fields(cls)
        },
        kw_only=True,
        frozen=True,
    )
    out.__doc__ = cls.__doc__.replace(":class:", "")
    return out


_AstroOptions = _param_cls_factory(AstroOptions)
_AstroParams = _param_cls_factory(AstroParams)
_CosmoParams = _param_cls_factory(CosmoParams)
_MatterOptions = _param_cls_factory(MatterOptions)
_SimulationOptions = _param_cls_factory(SimulationOptions)


@Parameter(name="*")
@dataclass
class Parameters:
    """A trimmed-down version of InputParameters with all defaults of None."""

    simulation_options: Annotated[
        _SimulationOptions, Parameter(name="*", group="SimulationOptions")
    ] = _SimulationOptions()
    astro_options: Annotated[
        _AstroOptions, Parameter(name="*", group="AstroOptions")
    ] = _AstroOptions()
    astro_params: Annotated[_AstroParams, Parameter(name="*", group="AstroParams")] = (
        _AstroParams()
    )
    cosmo_params: Annotated[_CosmoParams, Parameter(name="*", group="CosmoParams")] = (
        _CosmoParams()
    )
    matter_options: Annotated[
        _MatterOptions, Parameter(name="*", group="MatterOptions")
    ] = _MatterOptions()


def _get_inputs(
    options: RunParams | ParameterSelection, params: Parameters
) -> tuple[InputParameters, bool]:
    pselect = options.param_selection if isinstance(options, RunParams) else options

    seed = getattr(options, "seed", 42)

    # Turn our dummy Parameters into dictionaries, ignoring any unset values (None)
    kwargs = {}
    for _field in fields(Parameters):
        this = getattr(params, _field.name)
        kwargs |= {
            name: val for name, val in attrs.asdict(this).items() if val is not None
        }

    inputs = InputParameters.from_template(
        pselect.param_file or pselect.template, random_seed=seed, **kwargs
    )

    modified = False
    if kwargs:
        # Other functions need to know if we modified the template/file params at all,
        # so we make a version without the changes to compare.
        without_kw = InputParameters.from_template(
            pselect.param_file or pselect.template, random_seed=seed
        )
        modified = without_kw != inputs

    return inputs, modified


@cfg.command(name="avail")
def cfg_avail():
    """Print all available builtin templates."""
    templates = list_templates()
    for template in templates:
        name = template["name"]

        cns.print(f"[bold purple]{name}[/bold purple]", end=" ")
        aliases = template["aliases"]
        aliases.remove(name)  # Remove the name from aliases if present

        if aliases:
            cns.print(f"[purple] | {' | '.join(aliases)}")
        else:
            cns.print()

        cns.print(f"\t{template['description']}")


def pretty_print_inputs(
    inputs: InputParameters, name: str, description: str, mode: TOMLMode = "full"
):
    inputs_dct = convert_inputs_to_dict(inputs, mode, only_cstruct_params=True)

    @group()
    def get_panel_elements():
        yield description + "\n"

        for structname, params in inputs_dct.items():
            if params:
                yield Rule(f"[bold purple]{structname}")

                keys = [f"[bold]{k}[/bold]:" for k in params]
                vals = [
                    f"{f'[green italic]{v}' if isinstance(v, str) else v}"
                    for v in params.values()
                ]
                yield Columns(
                    [val for pair in zip(keys, vals, strict=False) for val in pair],
                    expand=False,
                )
                yield ""

    cns.print(Panel(get_panel_elements(), title=name, highlight=True))


@cfg.command(name="show")
def cfg_show(
    names: Annotated[
        list[Literal[*AVAILABLE_TEMPLATES]], Parameter(accepts_keys=False)
    ],
    mode: TOMLMode = "full",
):
    """Show and describe an in-built template."""
    template_desc = [
        next(
            f"[blue bold]{t['name']}[/]: {t['description']}"
            for t in list_templates()
            if name.upper() in (tt.upper() for tt in t["aliases"])
        )
        for name in names
    ]

    inputs = InputParameters.from_template(names, random_seed=0)
    pretty_print_inputs(
        inputs=inputs,
        name=" + ".join(names),
        description="\n".join(template_desc),
        mode=mode,
    )


@cfg.command(name="create")
def template_create(
    out: Annotated[
        cyctp.TomlPath,
        Parameter(validator=(vld.Path(file_okay=False, dir_okay=False, ext=("toml",)))),
    ],
    param_selection: ParameterSelection = ParameterSelection(),
    user_params: Parameters = Parameters(),
    mode: TOMLMode = "full",
):
    """Create a new full simulation parameter template.

    The created template file (TOML file) contains *all* of the available parameters.
    To create it, use a base template (either via --param-file or --template) and
    optionally override any particular simulation parameters. To see the available
    simulation parameters and how to specify them, use `21cmfast run params --help`.
    """
    inputs, _ = _get_inputs(param_selection, user_params)
    if not out.parent.exists():
        out.parent.mkdir(exist_ok=True, parents=True)

    pretty_print_inputs(
        inputs=inputs,
        name="Custom Template",
        description="The minimal set of non-default parameters of your custom model",
        mode="minimal",
    )
    write_template(inputs, out, mode=mode)
    cns.print(f":duck:[spring_green3] Wrote new template file at [purple]{out}")


def _run_setup(
    options: RunParams,
    params: Parameters,
    zmin: float | None = None,
    force_nodez: bool = False,
) -> InputParameters:
    print_banner()

    cns.print(Rule("Setting Up The Simulation", characters="="))

    def custom_rich_warning(message: str, *args, **kwargs):
        cns.print(f"[orange1]:warning: {message}")

    cvlmodule._console = cns
    warnings.showwarning = custom_rich_warning
    logger.setLevel(options.verbosity)

    inputs, modified = _get_inputs(options, params)

    if zmin is not None and (inputs.evolution_required or force_nodez):
        inputs = inputs.with_logspaced_redshifts(zmin=zmin)

    if (
        modified
        or options.param_selection.param_file is None
        or len(options.param_selection.param_file) > 1
        or (len(options.param_selection.template) > 1)
        or options.outcfg is not None
    ):
        if options.outcfg is not None:
            config_file = options.outcfg
        elif not modified and options.param_selection.param_file is None:
            name = "_and_".join(options.param_selection.template)
            config_file = options.cachedir / f"{name}.toml"
        else:
            config_file = options.cachedir / f"config-{uuid.uuid4().hex[:6]}.toml"

        write_template(inputs, config_file)
        cns.print(
            f":duck: [spring_green3]Wrote full configuration to [purple]{config_file}"
        )
        name = config_file.name
    else:
        name = " + ".join(cfg.name for cfg in options.param_selection.param_file)

    pretty_print_inputs(
        inputs=inputs,
        name=name,
        description="The minimal set of non-default parameters of your model",
        mode="minimal",
    )

    cns.print()
    cns.print(Rule("Starting Simulation", characters="="))
    return inputs


@run.command
def params(params: Parameters = Parameters()):
    """Show the current simulation parameters."""
    cns.print("Usage: 21cmfast run params --help")


@run.command()
def ics(
    options: RunParams = RunParams(),
    params: Annotated[Parameters, Parameter(show=False, name="*")] = Parameters(),
    min_evolved_redshift: Annotated[
        float, Parameter(name=("--zmin-evolution", "--zmin"))
    ] = 5.5,
):
    """Run a single iteration of 21cmFAST init, saving results to file.

    To specify simulation parameters, use a base template (either via --param-file or
    --template) and optionally override any particular simulation parameters. To see the
    available simulation parameters and how to specify them, use
    `21cmfast run params --help`.

    Parameters
    ----------
    min_evolved_redshift
        The minimum redshift down to which to evolve the simulation. For some simulation
        configurations, this is not used at all, while for others it will subtly change
        the evolution.
    """
    inputs = _run_setup(options, params, zmin=min_evolved_redshift)
    cache = OutputCache(options.cachedir)
    rc = RunCache.from_inputs(inputs=inputs, cache=cache)

    if rc.InitialConditions.exists():
        if options.regenerate:
            cns.print(
                "[yellow]:warning: Initial conditions already exist, but regeneration is requested. Overriding."
            )
        else:
            cns.print(
                f"[spring_green3]:duck: Initial conditions already exist at [purple]'{rc.InitialConditions}'[/purple], skipping computation."
            )
            return

    compute_initial_conditions(
        inputs=inputs,
        regenerate=options.regenerate,
        write=True,
        cache=OutputCache(options.cachedir),
    )

    cns.print(
        f"[spring_green3]:duck: Saved initial conditions to {rc.InitialConditions}"
    )


@run.command()
def coeval(
    redshifts: Annotated[list[float], Parameter(alias=("-z",), required=True)],
    options: RunParams = RunParams(),
    params: Annotated[Parameters, Parameter(show=False, name="*")] = Parameters(),
    out: Annotated[cyctp.ExistingDirectory, Parameter(name=("--out", "-o"))] = Path(),
    save_all_redshifts: Annotated[
        bool, Parameter(name=("--save-all-redshifts", "-a", "--all"))
    ] = False,
    min_evolved_redshift: Annotated[
        float, Parameter(name=("--zmin-evolution", "--zmin"))
    ] = 5.5,
):
    """Generate coeval cubes at given redshifts.

    To specify simulation parameters, use a base template (either via --param-file or
    --template) and optionally override any particular simulation parameters. To see the
    available simulation parameters and how to specify them, use
    `21cmfast run params --help`.

    Parameters
    ----------
    redshifts
        The redshifts at which to generate the coeval boxes.
    out
        The path at which to save the coeval boxes. The coeval data at each
        redshift will be saved to the filename "coeval_z{redshift:.2f}.h5".
    save_all_redshifts
        Whether to save all redshifts in `node_redshifts` (i.e. all those
        in the evolution of the simulation), or only those in the redshifts given.
    min_evolved_redshift
        The minimum redshift down to which to evolve the simulation. For some simulation
        configurations, this is not used at all, while for others it will subtly change
        the evolution.
    """
    inputs = _run_setup(options, params, zmin=min_evolved_redshift)

    for coeval, in_outputs in generate_coeval(
        out_redshifts=redshifts,
        inputs=inputs,
        regenerate=options.regenerate,
        cache=OutputCache(options.cachedir),
        write=getattr(
            CacheConfig, options.cache_strategy.replace("dmfield", "noloop")
        )(),
        progressbar=options.progress,
    ):
        if not in_outputs and not save_all_redshifts:
            continue

        outfile = out / f"coeval_z{coeval.redshift:.2f}.h5"
        coeval.save(outfile)
        cns.print(
            f"[spring_green3]:duck: Saved z={coeval.redshift:.2f} coeval box to [purple]{outfile}."
        )


@run.command()
def lightcone(
    options: RunParams = RunParams(),
    params: Annotated[Parameters, Parameter(show=False, name="*")] = Parameters(),
    redshift_range: tuple[float, float] = (6.0, 30.0),
    out: Annotated[
        Path,
        Parameter(validator=(vld.Path(dir_okay=False, file_okay=False, ext=("h5",)),)),
    ] = Path("lightcone.h5"),
    lightcone_quantities: Annotated[tuple[str], Parameter(alias=("--lq",))] = (
        "brightness_temp",
    ),
    min_evolved_redshift: Annotated[
        float, Parameter(name=("--zmin-evolution", "--zmin"))
    ] = 5.5,
):
    """Generate a lightcone between given redshifts.

    To specify simulation parameters, use a base template (either via --param-file or
    --template) and optionally override any particular simulation parameters. To see the
    available simulation parameters and how to specify them, use
    `21cmfast run params --help`.

    Parameters
    ----------
    redshift_range
        The redshifts between which to generate the lightcone.
    out
        The filename to which to save the lightcone data.
    lightcone_quantities
        Computed fields to generate lightcones for.
    min_evolved_redshift
        The minimum redshift down to which to evolve the simulation. For some simulation
        configurations, this is not used at all, while for others it will subtly change
        the evolution.
    """
    if not out.parent.exists():
        out.parent.mkdir(parents=True, exist_ok=True)

    inputs = _run_setup(options, params, min_evolved_redshift, force_nodez=True)

    # For now, always use the old default lightconing algorithm
    lcn = RectilinearLightconer.between_redshifts(
        min_redshift=redshift_range[0],
        max_redshift=redshift_range[1],
        resolution=inputs.simulation_options.cell_size,
        cosmo=inputs.cosmo_params.cosmo,
        quantities=lightcone_quantities,
    )

    lc = run_lightcone(
        lightconer=lcn,
        inputs=inputs,
        regenerate=options.regenerate,
        write=getattr(
            CacheConfig, options.cache_strategy.replace("dmfield", "noloop")
        )(),
        cache=OutputCache(options.cachedir),
        progressbar=options.progress,
    )

    lc.save(out)

    cns.print(f"[spring_green3]:duck: Saved Lightcone to {out}.")


@dev.command(name="feature")
def pr_feature(
    options: RunParams = RunParams(),
    params: Parameters = Parameters(),
    redshift_range: tuple[float, float] = (6.0, 30.0),
    outdir: cyctp.ExistingDirectory = Path(),
):
    """
    Create standard plots comparing a default simulation against a simulation with a new feature.

    The base of the comparison is set by either --param-file or --template.
    The "new feature" which is to be compared is set by overriding specific parameters
    at the command line (e.g --use-ts-fluct).

    Plots are saved in the current directory, with the prefix "pr_feature".

    Parameters
    ----------
    redshift_range
        The redshifts between which to compute the lightcones for comparison.
    """
    import powerbox

    inputs_default, _ = _get_inputs(options, Parameters())
    inputs_new, _ = _get_inputs(options, params)

    inputs_default = inputs_default.with_logspaced_redshifts(
        zmin=redshift_range[0] - 0.1
    )
    inputs_new = inputs_new.with_logspaced_redshifts(zmin=redshift_range[0] - 0.1)

    # For now, always use the old default lightconing algorithm
    lcn = RectilinearLightconer.between_redshifts(
        min_redshift=redshift_range[0],
        max_redshift=redshift_range[1],
        resolution=inputs_default.simulation_options.cell_size,
        cosmo=inputs_default.cosmo_params.cosmo,
        quantities=("brightness_temp",),
    )

    cns.print("Running default lightcone...")

    lc_default = run_lightcone(
        lightconer=lcn,
        inputs=inputs_default,
        regenerate=options.regenerate,
        write=getattr(
            CacheConfig, options.cache_strategy.replace("dmfield", "noloop")
        )(),
        cache=OutputCache(options.cachedir),
        progressbar=True,
    )

    cns.print("Running lightcone with new feature...")
    lc_new = run_lightcone(
        lightconer=lcn,
        inputs=inputs_new,
        regenerate=options.regenerate,
        write=getattr(
            CacheConfig, options.cache_strategy.replace("dmfield", "noloop")
        )(),
        cache=OutputCache(options.cachedir),
        progressbar=True,
    )

    cns.print("Plotting lightcone slices...")
    fig, ax = plt.subplots(3, 1, sharex=True, sharey=True)

    vmin = -150
    vmax = 30

    plotting.lightcone_sliceplot(lc_default, ax=ax[0], fig=fig, vmin=vmin, vmax=vmax)
    ax[0].set_title("Default")

    plotting.lightcone_sliceplot(
        lc_new, ax=ax[1], fig=fig, cbar=False, vmin=vmin, vmax=vmax
    )
    ax[1].set_title("New")

    plotting.lightcone_sliceplot(
        lc_default, lightcone2=lc_new, cmap="bwr", ax=ax[2], fig=fig
    )
    ax[2].set_title("Difference")

    plt.savefig(f"{outdir}/pr_feature_lightcone_2d_brightness_temp.pdf")

    def rms(x, axis=None):
        return np.sqrt(np.mean(x**2, axis=axis))

    cns.print("Plotting lightcone history...")
    fig, ax = plt.subplots(4, 1, sharex=True, gridspec_kw={"hspace": 0.05})
    ax[0].plot(
        lc_default.inputs.node_redshifts,
        lc_default.global_quantities["neutral_fraction"],
        label="Default",
    )
    ax[0].plot(
        lc_new.inputs.node_redshifts,
        lc_new.global_quantities["neutral_fraction"],
        label="New",
    )
    ax[0].set_ylabel(r"$x_{\rm HI}$")
    ax[0].legend()

    ax[1].plot(
        lc_default.inputs.node_redshifts,
        lc_default.global_quantities["brightness_temp"],
        label="Default",
    )
    ax[1].plot(
        lc_new.inputs.node_redshifts,
        lc_new.global_quantities["brightness_temp"],
        label="New",
    )
    ax[1].set_ylabel("$T_b$ [K]")
    ax[3].set_xlabel("z")

    rms_diff = rms(lc_default.lightcones["brightness_temp"], axis=(0, 1)) - rms(
        lc_new.lightcones["brightness_temp"], axis=(0, 1)
    )
    ax[2].plot(lc_default.lightcone_redshifts, rms_diff, label="RMS")
    ax[2].plot(
        lc_new.inputs.node_redshifts,
        lc_default.global_quantities["neutral_fraction"]
        - lc_new.global_quantities["neutral_fraction"],
        label="$x_{HI}$",
    )
    ax[2].plot(
        lc_new.inputs.node_redshifts,
        lc_default.global_quantities["brightness_temp"]
        - lc_new.global_quantities["brightness_temp"],
        label="$T_b$",
    )
    ax[2].legend()
    ax[2].set_ylabel("Differences")

    diff_rms = rms(
        lc_default.lightcones["brightness_temp"] - lc_new.lightcones["brightness_temp"],
        axis=(0, 1),
    )
    ax[3].plot(lc_default.lightcone_redshifts, diff_rms)
    ax[3].set_ylabel("RMS of Diff.")

    plt.savefig(f"{outdir}/pr_feature_history.pdf")

    cns.print("Plotting power spectra history...")
    p_default = []
    p_new = []
    z = []
    thickness = 200  # Mpc
    ncells = int(thickness / lc_new.cell_size)
    chunk_size = lc_new.cell_size * ncells
    start = 0
    cns.print(ncells)
    while start + ncells <= lc_new.shape[-1]:
        pd, k = powerbox.get_power(
            lc_default.lightcones["brightness_temp"][:, :, start : start + ncells],
            (*lc_default.lightcone_dimensions[:2], chunk_size),
            bins_upto_boxlen=True,
        )
        p_default.append(pd)

        pn, k = powerbox.get_power(
            lc_new.lightcones["brightness_temp"][:, :, start : start + ncells],
            (*lc_new.lightcone_dimensions[:2], chunk_size),
            bins_upto_boxlen=True,
        )
        p_new.append(pn)
        z.append(lc_new.lightcone_redshifts[start])

        start += ncells

    p_default = np.array(p_default).T
    p_new = np.array(p_new).T

    fig, ax = plt.subplots(2, 1, sharex=True)
    ax[0].set_yscale("log")

    inds = [
        np.where(np.abs(k - 0.1) == np.abs(k - 0.1).min())[0][0],
        np.where(np.abs(k - 0.2) == np.abs(k - 0.2).min())[0][0],
        np.where(np.abs(k - 0.5) == np.abs(k - 0.5).min())[0][0],
        np.where(np.abs(k - 1) == np.abs(k - 1).min())[0][0],
    ]

    for i, (pdef, pnew, kk) in enumerate(
        zip(p_default[inds], p_new[inds], k[inds], strict=True)
    ):
        ax[0].plot(z, pdef, ls="--", label=f"k={kk:.2f}", color=f"C{i}")
        ax[0].plot(z, pnew, ls="-", color=f"C{i}")
        ax[1].plot(z, np.log10(pdef / pnew), ls="-", color=f"C{i}")
    ax[1].set_xlabel("z")
    ax[0].set_ylabel(r"$\Delta^2 [{\rm mK}^2]$")
    ax[1].set_ylabel(r"log ratio of $\Delta^2 [{\rm mK}^2]$")
    ax[0].legend()

    plt.savefig(f"{outdir}/pr_feature_power_history.pdf")


@pred.command(name="struct-size")
def predict_struct_size(
    param_selection: ParameterSelection = ParameterSelection(),
    user_params: Parameters = Parameters(),
    unit: Literal["b", "kb", "mb", "gb"] | None = None,
    cache_config: Literal["on", "off", "noloop", "last_step_only"] = "on",
):
    """Compute the required storage per output kind for given inputs."""
    from .management import get_expected_sizes

    inputs, _ = _get_inputs(param_selection, user_params)
    sizes = get_expected_sizes(
        inputs, cache_config=getattr(CacheConfig, cache_config)()
    )
    if len(sizes) == 0:
        cns.print("No output structs are expected for given inputs and cache-config.")
        return

    units = ["b", "kb", "mb", "gb", "tb"]
    if unit is None:
        bigness = int((np.log(list(sizes.values())) / np.log(1024)).max())
    else:
        bigness = units.index(unit)

    table = Table(title="Output Struct Sizes")
    table.add_column("Struct Name")
    table.add_column("Size")
    table.add_column("Unit")

    for name, size in sizes.items():
        table.add_row(name, f"{size / 1024**bigness:.2f}", units[bigness].upper())
    table.add_section()
    table.add_row(
        "Total", f"{sum(sizes.values()) / 1024**bigness:.2f}", units[bigness].upper()
    )
    cns.print(table)


@pred.command(name="storage-size")
def predict_storage_size(
    param_selection: ParameterSelection = ParameterSelection(),
    user_params: Parameters = Parameters(),
    min_evolved_redshift: Annotated[
        float, Parameter(name=("--zmin-evolution", "--zmin"))
    ] = 5.5,
    unit: Literal["b", "kb", "mb", "gb"] | None = None,
    cache_config: Literal["on", "off", "noloop", "last_step_only"] = "on",
):
    """Compute the required storage for an entire run."""
    from .management import get_total_storage_size

    inputs, _ = _get_inputs(param_selection, user_params)
    inputs = inputs.with_logspaced_redshifts(zmin=min_evolved_redshift)

    sizes = get_total_storage_size(
        inputs, cache_config=getattr(CacheConfig, cache_config)()
    )

    if len(sizes) == 0:
        cns.print("No output structs are expected for given inputs and cache-config.")
        return

    units = ["b", "kb", "mb", "gb", "tb"]
    if unit is None:
        bigness = int(np.log(max(size for _, size in sizes.values())) / np.log(1024))
    else:
        bigness = units.index(unit)

    table = Table(title="Storage Sizes")
    table.add_column("Struct Name")
    table.add_column(f"Size [{units[bigness].upper()}]")
    table.add_column("Quantity")

    total_size = 0
    total_quant = 0
    for name, (quant, size) in sizes.items():
        total_size += size
        total_quant += quant
        table.add_row(name, f"{size / 1024**bigness:.2f}", f"{quant}")
    table.add_section()
    table.add_row("Total", f"{total_size / 1024**bigness:.2f}", f"{total_quant}")
    cns.print(table)


if __name__ == "__main__":
    app()



"""Low-level python wrappers of C functions."""

import logging
from collections.abc import Callable, Sequence
from functools import cache
from typing import Literal

import numpy as np
from numpy.typing import NDArray
from scipy.interpolate import interp1d

from .._cfg import config
from ..c_21cmfast import ffi, lib
from ._utils import _process_exitcode
from .inputs import (
    InputParameters,
)

logger = logging.getLogger(__name__)

# Ideally, backend functions that we access here should do all the broadcasting/initialisation themselves
# These decorators are for lower functions which are called directly in one or two lines, like delta_crit

# TODO: a lot of these assume input as numpy arrays via use of .shape, explicitly require this


def broadcast_input_struct(inputs: InputParameters):
    """Broadcast the parameters to the C library."""
    lib.Broadcast_struct_global_all(
        inputs.simulation_options.cstruct,
        inputs.matter_options.cstruct,
        inputs.cosmo_params.cstruct,
        inputs.astro_params.cstruct,
        inputs.astro_options.cstruct,
        inputs.cosmo_tables.cstruct,
    )


def free_cosmo_tables():
    """Free the memory of cosmo_tables_global that was allocated at the C backend."""
    lib.Free_cosmo_tables_global()


def broadcast_params(func: Callable) -> Callable:
    """Broadcast the parameters to the C library before calling the function.

    This should be added as a decorator to any function which accesses the
    21cmFAST if it does not directly call `broadcast_input_struct`.
    """

    def wrapper(*args, inputs: InputParameters, **kwargs):
        broadcast_input_struct(inputs)
        try:
            out = func(*args, inputs=inputs, **kwargs)
        except:
            # Always free on error, regardless of the flag
            free_cosmo_tables()
            raise  # Re-raise the original exception
        else:
            # Only free on success if the flag allows it
            if kwargs.get("free_cosmo_tables", True):
                free_cosmo_tables()
            return out

    return wrapper


def init_backend_ps(func: Callable) -> Callable:
    """Initialise the backend power-spectrum before calling the function.

    This should be added as a decorator to any function which uses the cosmology
    from the 21cmFAST backend without passing through our regular functions.
    """

    @broadcast_params
    def wrapper(*args, **kwargs):
        lib.init_ps()
        return func(*args, **kwargs)

    return wrapper


def init_sigma_table(func: Callable) -> Callable:
    """Initialise the the sigma interpolation table before calling the function.

    This should be added as a decorator to any function which calls lib.EvaluateSigma
    or the sigma tables directly.
    """

    @init_backend_ps
    def wrapper(*args, inputs: InputParameters, **kwargs):
        sigma_min_mass = kwargs.get("M_min", 1e5)
        sigma_max_mass = kwargs.get("M_max", 1e16)
        if inputs.matter_options.USE_INTERPOLATION_TABLES != "no-interpolation":
            lib.initialiseSigmaMInterpTable(sigma_min_mass, sigma_max_mass)
        return func(*args, inputs=inputs, **kwargs)

    return wrapper


def init_gl(func: Callable) -> Callable:
    """Initialise the Gauss-Legendre integration if required before calling the function.

    Calculates the abcissae weights and stores them as arrays in the backend when either of the
    HMF integrals is set to use Gauss-Legendre integration. This should be added as a decorator to
    any function which calls backend integrals directly.
    """

    @init_sigma_table
    def wrapper(*args, inputs: InputParameters, **kwargs):
        if "GAUSS-LEGENDRE" in (
            inputs.astro_options.INTEGRATION_METHOD_ATOMIC,
            inputs.astro_options.INTEGRATION_METHOD_MINI,
        ):
            # no defualt since GL mass limits are strict
            lib.initialise_GL(np.log(kwargs.get("M_min")), np.log(kwargs.get("M_max")))
        return func(*args, inputs=inputs, **kwargs)

    return wrapper


@broadcast_params
def get_expected_nhalo(*, redshift: float, inputs: InputParameters, **kwargs) -> int:
    """Get the expected number of halos in a given box.

    Parameters
    ----------
    redshift : float
        The redshift at which to calculate the halo list.
    inputs: :class:`~InputParameters`
        The input parameters of the run
    """
    return lib.expected_nhalo(
        redshift,
    )


@broadcast_params
def get_halo_catalog_buffer_size(
    *, redshift: float, inputs: InputParameters, min_size: int = 1000000, **kwargs
) -> int:
    """Compute the required size of the memory buffer to hold a halo list.

    Parameters
    ----------
    redshift : float
        The redshift at which to calculate the halo list.
    inputs: :class:`~InputParameters`
        The input parameters of the run
    min_size : int, optional
        A minimum size to be used as the buffer.
    """
    # find the buffer size from expected halos in the box
    hbuffer_size = get_expected_nhalo(
        redshift=redshift,
        inputs=inputs,
        free_cosmo_tables=kwargs.get("free_cosmo_tables", True),
    )
    hbuffer_size = int((hbuffer_size + 1) * config["HALO_CATALOG_MEM_FACTOR"])

    # set a minimum in case of fluctuation at high z
    return int(max(hbuffer_size, min_size))


@broadcast_params
def compute_tau(
    *,
    redshifts: Sequence[float],
    global_xHI: Sequence[float],
    inputs: InputParameters,
    z_re_HeII: float = 3.0,
) -> float:
    """Compute the optical depth to reionization under the given model.

    Parameters
    ----------
    redshifts : array-like
        Redshifts defining an evolution of the neutral fraction.
    global_xHI : array-like
        The mean neutral fraction at `redshifts`.
    inputs : :class:`~InputParameters`
        Defines the input parameters of the run
    z_re_HeII : float, optional
        The redshift at which helium reionization occurs.

    Returns
    -------
    tau : float
        The optical depth to reionization

    Raises
    ------
    ValueError :
        If `redshifts` and `global_xHI` have inconsistent length or if redshifts are not
        in ascending order.
    """
    if len(redshifts) != len(global_xHI):
        raise ValueError("redshifts and global_xHI must have same length")

    if not np.all(np.diff(redshifts) > 0):
        raise ValueError("redshifts and global_xHI must be in ascending order")

    # Convert the data to the right type
    redshifts = np.array(redshifts, dtype="float32")
    global_xHI = np.array(global_xHI, dtype="float32")

    z = ffi.cast("float *", ffi.from_buffer(redshifts))
    xHI = ffi.cast("float *", ffi.from_buffer(global_xHI))

    # Run the C code
    return lib.ComputeTau(
        len(redshifts),
        z,
        xHI,
        z_re_HeII,
    )


@broadcast_params
def compute_luminosity_function(
    *,
    redshifts: Sequence[float],
    inputs: InputParameters,
    nbins: int = 100,
    mturnovers: np.ndarray | None = None,
    mturnovers_mini: np.ndarray | None = None,
    component: Literal["both", "acg", "mcg"] = "both",
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Compute a the luminosity function over a given number of bins and redshifts.

    Parameters
    ----------
    redshifts : array-like
        The redshifts at which to compute the luminosity function.
    inputs: :class:`~InputParameters`
        The input parameters defining the simulation run.
    nbins : int, optional
        The number of luminosity bins to produce for the luminosity function.
    mturnovers : array-like, optional
        The turnover mass at each redshift for massive halos (ACGs).
        Only required when USE_MINI_HALOS is True.
    mturnovers_mini : array-like, optional
        The turnover mass at each redshift for minihalos (MCGs).
        Only required when USE_MINI_HALOS is True.
    component : str, {'both', 'acg', 'mcg}
        The component of the LF to be calculated.

    Returns
    -------
    Muvfunc : np.ndarray
        Magnitude array (i.e. brightness). Shape [nredshifts, nbins]
    Mhfunc : np.ndarray
        Halo mass array. Shape [nredshifts, nbins]
    lfunc : np.ndarray
        Number density of haloes corresponding to each bin defined by `Muvfunc`.
        Shape [nredshifts, nbins].
    """
    astro_options = inputs.astro_options
    astro_params = inputs.astro_params

    redshifts = np.array(redshifts, dtype="float32")
    if astro_options.USE_MINI_HALOS:
        if component in ["both", "acg"]:
            if mturnovers is None:
                raise ValueError(
                    "calculating ACG LFs with mini-halo feature requires users to "
                    "specify mturnovers!"
                )

            mturnovers = np.array(mturnovers, dtype=np.float32)
            if len(mturnovers) != len(redshifts):
                raise ValueError(
                    f"mturnovers ({len(mturnovers)}) does not match the length of "
                    f"redshifts ({len(redshifts)})"
                )
        if component in ["both", "mcg"]:
            if mturnovers_mini is None:
                raise ValueError(
                    "calculating MCG LFs with mini-halo feature requires users to "
                    "specify mturnovers_MINI!"
                )

            mturnovers_mini = np.array(mturnovers_mini, dtype="float32")
            if len(mturnovers_mini) != len(redshifts):
                raise ValueError(
                    f"mturnovers_MINI ({len(mturnovers)}) does not match the length of "
                    f"redshifts ({len(redshifts)})"
                )

    else:
        mturnovers = (
            np.zeros(len(redshifts), dtype=np.float32) + 10**astro_params.M_TURN
        )
        component = "acg"

    lfunc = np.zeros(len(redshifts) * nbins)
    Muvfunc = np.zeros(len(redshifts) * nbins)
    Mhfunc = np.zeros(len(redshifts) * nbins)

    lfunc.shape = (len(redshifts), nbins)
    Muvfunc.shape = (len(redshifts), nbins)
    Mhfunc.shape = (len(redshifts), nbins)

    c_Muvfunc = ffi.cast("double *", ffi.from_buffer(Muvfunc))
    c_Mhfunc = ffi.cast("double *", ffi.from_buffer(Mhfunc))
    c_lfunc = ffi.cast("double *", ffi.from_buffer(lfunc))

    lfunc_MINI = np.zeros(len(redshifts) * nbins)
    Muvfunc_MINI = np.zeros(len(redshifts) * nbins)
    Mhfunc_MINI = np.zeros(len(redshifts) * nbins)

    lfunc_MINI.shape = (len(redshifts), nbins)
    Muvfunc_MINI.shape = (len(redshifts), nbins)
    Mhfunc_MINI.shape = (len(redshifts), nbins)

    c_Muvfunc_MINI = ffi.cast("double *", ffi.from_buffer(Muvfunc_MINI))
    c_Mhfunc_MINI = ffi.cast("double *", ffi.from_buffer(Mhfunc_MINI))
    c_lfunc_MINI = ffi.cast("double *", ffi.from_buffer(lfunc_MINI))

    if component in ("both", "acg"):
        # Run the C code
        errcode = lib.ComputeLF(
            nbins,
            1,
            len(redshifts),
            ffi.cast("float *", ffi.from_buffer(redshifts)),
            ffi.cast("float *", ffi.from_buffer(mturnovers)),
            c_Muvfunc,
            c_Mhfunc,
            c_lfunc,
        )

        _process_exitcode(
            errcode,
            lib.ComputeLF,
            (
                nbins,
                1,
                len(redshifts),
            ),
        )

    if component in ("both", "mcg"):
        # Run the C code
        errcode = lib.ComputeLF(
            nbins,
            2,
            len(redshifts),
            ffi.cast("float *", ffi.from_buffer(redshifts)),
            ffi.cast("float *", ffi.from_buffer(mturnovers_mini)),
            c_Muvfunc_MINI,
            c_Mhfunc_MINI,
            c_lfunc_MINI,
        )

        _process_exitcode(
            errcode,
            lib.ComputeLF,
            (
                nbins,
                2,
                len(redshifts),
            ),
        )

    if component == "both":
        # redo the Muv range using the faintest (most likely MINI) and the brightest (most likely massive)
        lfunc_all = np.zeros(len(redshifts) * nbins)
        Muvfunc_all = np.zeros(len(redshifts) * nbins)
        Mhfunc_all = np.zeros(len(redshifts) * nbins * 2)

        lfunc_all.shape = (len(redshifts), nbins)
        Muvfunc_all.shape = (len(redshifts), nbins)
        Mhfunc_all.shape = (len(redshifts), nbins, 2)

        for iz in range(len(redshifts)):
            Muvfunc_all[iz] = np.linspace(
                np.min([Muvfunc.min(), Muvfunc_MINI.min()]),
                np.max([Muvfunc.max(), Muvfunc_MINI.max()]),
                nbins,
            )
            lfunc_all[iz] = np.log10(
                10
                ** (
                    interp1d(Muvfunc[iz], lfunc[iz], fill_value="extrapolate")(
                        Muvfunc_all[iz]
                    )
                )
                + 10
                ** (
                    interp1d(
                        Muvfunc_MINI[iz], lfunc_MINI[iz], fill_value="extrapolate"
                    )(Muvfunc_all[iz])
                )
            )
            Mhfunc_all[iz] = np.array(
                [
                    interp1d(Muvfunc[iz], Mhfunc[iz], fill_value="extrapolate")(
                        Muvfunc_all[iz]
                    ),
                    interp1d(
                        Muvfunc_MINI[iz], Mhfunc_MINI[iz], fill_value="extrapolate"
                    )(Muvfunc_all[iz]),
                ],
            ).T
        lfunc_all[lfunc_all <= -30] = np.nan
        return Muvfunc_all, Mhfunc_all, lfunc_all
    elif component == "acg":
        lfunc[lfunc <= -30] = np.nan
        return Muvfunc, Mhfunc, lfunc
    elif component == "mcg":
        lfunc_MINI[lfunc_MINI <= -30] = np.nan
        return Muvfunc_MINI, Mhfunc_MINI, lfunc_MINI
    else:
        raise ValueError(
            f"Unknown component '{component}'. Must be 'both', 'acg' or 'mcg'"
        )


@cache
def construct_fftw_wisdoms(
    *,
    use_fftw_wisdom: bool,
) -> int:
    """Construct all necessary FFTW wisdoms.

    Parameters
    ----------
    USE_FFTW_WISDOM : bool
        Whether we are interested in having FFTW wisdoms.

    """
    # Run the C code
    if use_fftw_wisdom:
        return lib.CreateFFTWWisdoms()
    else:
        return 0


@init_backend_ps
def get_matter_power_values(
    *,
    inputs: InputParameters,
    k_values: Sequence[float],
):
    """Evaluate the matter density power spectrum (at z=0) at a certain scale from the 21cmFAST backend."""
    return np.vectorize(lib.power_in_k)(k_values)


@init_backend_ps
def get_vcb_power_values(
    *,
    inputs: InputParameters,
    k_values: Sequence[float],
):
    """Evaluate the vcb power spectrum (at kinematic decoupling) at a certain scale from the 21cmFAST backend."""
    if inputs.matter_options.USE_RELATIVE_VELOCITIES:
        return np.vectorize(lib.power_in_vcb)(k_values)
    else:
        raise ValueError(
            "inputs.matter_options.USE_RELATIVE_VELOCITIES must be True in order to compute the v_cb power spectrum."
        )


@broadcast_params
def evaluate_sigma(
    *,
    inputs: InputParameters,
    masses: NDArray[float],
):
    """
    Evaluate the variance of a mass scale.

    Uses the 21cmfast backend
    """
    masses = masses.astype("f8")
    sigma = np.zeros_like(masses)
    dsigmasq = np.zeros_like(masses)

    lib.get_sigma(
        masses.size,
        ffi.cast("double *", ffi.from_buffer(masses)),
        ffi.cast("double *", ffi.from_buffer(sigma)),
        ffi.cast("double *", ffi.from_buffer(dsigmasq)),
    )

    return sigma, dsigmasq


@init_backend_ps
def get_growth_factor(
    *,
    inputs: InputParameters,
    redshift: float,
):
    """Get the growth factor at a given redshift."""
    return lib.dicke(redshift)


def get_condition_mass(inputs: InputParameters, R: float):
    """Determine condition masses for backend routines.

    Returns either mass contained within a radius,
    or mass of the Lagrangian cell on HII_DIM
    """
    rhocrit = (
        inputs.cosmo_params.cosmo.critical_density(0).to("M_sun Mpc-3").value
        * inputs.cosmo_params.OMm
    )
    if R == "cell":
        volume = (
            inputs.simulation_options.BOX_LEN / inputs.simulation_options.HII_DIM
        ) ** 3
    else:
        volume = 4.0 / 3.0 * np.pi * R**3

    return volume * rhocrit


@broadcast_params
def get_delta_crit(*, inputs: InputParameters, mass: float, redshift: float):
    """Get the critical collapse density given a mass, redshift and parameters."""
    sigma, _ = evaluate_sigma(inputs=inputs, masses=np.array([mass]))
    growth = get_growth_factor(inputs=inputs, redshift=redshift)
    return get_delta_crit_nu(inputs.matter_options.cdict["HMF"], sigma[0], growth)


def get_delta_crit_nu(hmf_int_flag: int, sigma: float, growth: float):
    """Get the critical density from sigma and growth factor."""
    # None of the parameter structs are used in this function so we don't need a broadcast
    return lib.get_delta_crit(hmf_int_flag, sigma, growth)


@broadcast_params
def evaluate_condition_integrals(
    inputs: InputParameters,
    cond_array: NDArray[float],
    redshift: float,
    redshift_prev: float | None = None,
):
    """Get the expected number and mass of halos given a condition.

    If USE_INTERPOLATION_TABLES is set to 'hmf-interpolation': Will crash if the table
    has not been initialised, only `cond_array` is used,
    and the rest of the arguments are taken from when the table was initialised.
    """
    cond_array = cond_array.astype("f8")
    n_halo = np.zeros_like(cond_array)
    m_coll = np.zeros_like(cond_array)

    lib.get_condition_integrals(
        redshift,
        redshift_prev if redshift_prev is not None else -1,
        cond_array.size,
        ffi.cast("double *", ffi.from_buffer(cond_array)),
        ffi.cast("double *", ffi.from_buffer(n_halo)),
        ffi.cast("double *", ffi.from_buffer(m_coll)),
    )

    return n_halo, m_coll


@broadcast_params
def integrate_chmf_interval(
    inputs: InputParameters,
    redshift: float,
    lnm_lower: NDArray[float],
    lnm_upper: NDArray[float],
    cond_values: NDArray[float],
    redshift_prev: float | None = None,
):
    """Evaluate conditional mass function integrals at a range of mass intervals."""
    if lnm_lower.shape != lnm_upper.shape:
        raise ValueError("the shapes of the two mass-limit arrays must be equal")
    assert np.all(lnm_lower < lnm_upper)

    out_prob = np.zeros((len(cond_values), len(lnm_lower)), dtype="f8")
    cond_values = cond_values.astype("f8")
    lnm_lower = lnm_lower.astype("f8")
    lnm_upper = lnm_upper.astype("f8")

    lib.get_halo_chmf_interval(
        redshift,
        redshift_prev if redshift_prev is not None else -1,
        len(cond_values),
        ffi.cast("double *", ffi.from_buffer(cond_values)),
        len(lnm_lower),
        ffi.cast("double *", ffi.from_buffer(lnm_lower)),
        ffi.cast("double *", ffi.from_buffer(lnm_upper)),
        ffi.cast("double *", ffi.from_buffer(out_prob)),
    )

    return out_prob


@broadcast_params
def evaluate_inverse_table(
    inputs: InputParameters,
    cond_array: NDArray[float],
    probabilities: NDArray[float],
    redshift: float,
    redshift_prev: float | None = None,
):
    """Get the expected number and mass of halos given a condition."""
    if cond_array.shape != probabilities.shape:
        raise ValueError(
            "the shapes of the input arrays `cond_array` and `probabilities"
            " must be equal."
        )

    if redshift_prev is None:
        redshift_prev = -1

    cond_array = cond_array.astype("f8")
    probabilities = probabilities.astype("f8")
    masses = np.zeros_like(cond_array)

    lib.get_halomass_at_probability(
        redshift,
        redshift_prev,
        cond_array.size,
        ffi.cast("double *", ffi.from_buffer(cond_array)),
        ffi.cast("double *", ffi.from_buffer(probabilities)),
        ffi.cast("double *", ffi.from_buffer(masses)),
    )

    return masses


@broadcast_params
def evaluate_FgtrM_cond(
    inputs: InputParameters,
    densities: NDArray[float],
    redshift: float,
    R: float,
):
    """Get the collapsed fraction from the backend, given a density and condition sigma."""
    densities = densities.astype("f8")
    fcoll = np.zeros_like(densities)
    dfcoll = np.zeros_like(densities)

    lib.get_conditional_FgtrM(
        redshift,
        R,
        densities.size,
        ffi.cast("double *", ffi.from_buffer(densities)),
        ffi.cast("double *", ffi.from_buffer(fcoll)),
        ffi.cast("double *", ffi.from_buffer(dfcoll)),
    )
    return fcoll, dfcoll


@broadcast_params
def evaluate_SFRD_z(
    *,
    inputs: InputParameters,
    redshifts: NDArray[float],
    log10mturns: NDArray[float],
):
    """Evaluate the global star formation rate density expected at a range of redshifts."""
    if redshifts.shape != log10mturns.shape:
        raise ValueError(
            f"the shapes of the input arrays `redshifts` {redshifts.shape} and `log10mturns` {log10mturns.shape}"
            " must be equal."
        )

    redshifts = redshifts.astype("f8")
    log10mturns = log10mturns.astype("f8")
    sfrd = np.zeros_like(redshifts)
    sfrd_mini = np.zeros_like(redshifts)

    lib.get_global_SFRD_z(
        redshifts.size,
        ffi.cast("double *", ffi.from_buffer(redshifts)),
        ffi.cast("double *", ffi.from_buffer(log10mturns)),
        ffi.cast("double *", ffi.from_buffer(sfrd)),
        ffi.cast("double *", ffi.from_buffer(sfrd_mini)),
    )

    return sfrd, sfrd_mini


@broadcast_params
def evaluate_Nion_z(
    *,
    inputs: InputParameters,
    redshifts: NDArray[float],
    log10mturns: NDArray[float],
):
    """Evaluate the global ionising emissivity expected at a range of redshifts."""
    if redshifts.shape != log10mturns.shape:
        raise ValueError(
            "the shapes of the input arrays `redshifts` and `log10mturns`"
            " must be equal."
        )

    redshifts = redshifts.astype("f8")
    log10mturns = log10mturns.astype("f8")
    nion = np.zeros_like(redshifts)
    nion_mini = np.zeros_like(redshifts)

    lib.get_global_Nion_z(
        redshifts.size,
        ffi.cast("double *", ffi.from_buffer(redshifts)),
        ffi.cast("double *", ffi.from_buffer(log10mturns)),
        ffi.cast("double *", ffi.from_buffer(nion)),
        ffi.cast("double *", ffi.from_buffer(nion_mini)),
    )

    return nion, nion_mini


@broadcast_params
def evaluate_SFRD_cond(
    *,
    inputs: InputParameters,
    redshift: float,
    radius: float,
    densities: NDArray[float],
    log10mturns: NDArray[float],
):
    """Evaluate the conditional star formation rate density expected at a range of densities."""
    if densities.shape != log10mturns.shape:
        raise ValueError(
            "the shapes of the input arrays `densities` and `log10mturns` must be equal"
        )

    densities = densities.astype("f8")
    log10mturns = log10mturns.astype("f8")
    sfrd = np.zeros_like(densities)
    sfrd_mini = np.zeros_like(densities)

    lib.get_conditional_SFRD(
        redshift,
        radius,
        densities.size,
        ffi.cast("double *", ffi.from_buffer(densities)),
        ffi.cast("double *", ffi.from_buffer(log10mturns)),
        ffi.cast("double *", ffi.from_buffer(sfrd)),
        ffi.cast("double *", ffi.from_buffer(sfrd_mini)),
    )

    return sfrd, sfrd_mini


@broadcast_params
def evaluate_Nion_cond(
    *,
    inputs: InputParameters,
    redshift: float,
    radius: float,
    densities: NDArray[float],
    l10mturns_acg: NDArray[float],
    l10mturns_mcg: NDArray[float],
):
    """Evaluate the conditional ionising emissivity expected at a range of densities."""
    if not (densities.shape == l10mturns_mcg.shape == l10mturns_acg.shape):
        raise ValueError(
            "the shapes of the input arrays `densities` and `log10mturns_x` must be equal"
        )

    densities = densities.astype("f8")
    l10mturns_acg = l10mturns_acg.astype("f8")
    l10mturns_mcg = l10mturns_mcg.astype("f8")
    nion = np.zeros_like(densities)
    nion_mini = np.zeros_like(densities)

    lib.get_conditional_Nion(
        redshift,
        radius,
        densities.size,
        ffi.cast("double *", ffi.from_buffer(densities)),
        ffi.cast("double *", ffi.from_buffer(l10mturns_acg)),
        ffi.cast("double *", ffi.from_buffer(l10mturns_mcg)),
        ffi.cast("double *", ffi.from_buffer(nion)),
        ffi.cast("double *", ffi.from_buffer(nion_mini)),
    )

    return nion, nion_mini


@broadcast_params
def evaluate_Xray_cond(
    *,
    inputs: InputParameters,
    redshift: float,
    radius: float,
    densities: NDArray[float],
    log10mturns: NDArray[float],
):
    """Evaluate the conditional star formation rate density expected at a range of densities."""
    if densities.shape != log10mturns.shape:
        raise ValueError(
            "the shapes of the input arrays `cond_array` and `probabilities"
            " must be equal."
        )

    densities = densities.astype("f8")
    log10mturns = log10mturns.astype("f8")
    xray = np.zeros_like(densities)

    lib.get_conditional_Xray(
        redshift,
        radius,
        densities.size,
        ffi.cast("double *", ffi.from_buffer(densities)),
        ffi.cast("double *", ffi.from_buffer(log10mturns)),
        ffi.cast("double *", ffi.from_buffer(xray)),
    )

    return xray


@broadcast_params
def sample_halos_from_conditions(
    *,
    inputs: InputParameters,
    redshift: float,
    cond_array,
    redshift_prev: float | None = None,
    buffer_size: int | None = None,
):
    """Construct a halo sample given a descendant catalogue and redshifts."""
    z_prev = -1 if redshift_prev is None else redshift_prev
    if buffer_size is None:
        buffer_size = get_halo_catalog_buffer_size(inputs=inputs, redshift=redshift)

    n_cond = cond_array.size
    # all coordinates zero
    crd_in = np.zeros(3 * n_cond).astype("f4")

    cond_array = cond_array.astype("f4")
    nhalo_out = np.zeros(1).astype("i4")
    N_out = np.zeros(n_cond).astype("i4")
    M_out = np.zeros(n_cond).astype("f8")
    exp_M = np.zeros(n_cond).astype("f8")
    exp_N = np.zeros(n_cond).astype("f8")
    halomass_out = np.zeros(buffer_size).astype("f4")
    halocrd_out = np.zeros(int(3 * buffer_size)).astype("i4")

    lib.single_test_sample(
        inputs.random_seed,
        n_cond,
        ffi.cast("float *", cond_array.ctypes.data),
        ffi.cast("float *", crd_in.ctypes.data),
        redshift,
        z_prev,
        ffi.cast("int *", nhalo_out.ctypes.data),
        ffi.cast("int *", N_out.ctypes.data),
        ffi.cast("double *", exp_N.ctypes.data),
        ffi.cast("double *", M_out.ctypes.data),
        ffi.cast("double *", exp_M.ctypes.data),
        ffi.cast("float *", halomass_out.ctypes.data),
        ffi.cast("float *", halocrd_out.ctypes.data),
    )

    return {
        "n_halo_total": nhalo_out[0],
        "halo_masses": halomass_out,
        "n_progenitors": N_out,
        "progenitor_mass": M_out,
        "expected_progenitors": exp_N,
        "expected_progenitor_mass": exp_M,
    }


@broadcast_params
def convert_halo_properties(
    *,
    redshift: float,
    inputs: InputParameters,
    halo_masses: NDArray[float],
    star_rng: NDArray[float],
    sfr_rng: NDArray[float],
    xray_rng: NDArray[float],
    halo_coords: NDArray[float] | None = None,
    vcb_grid: NDArray[float] | None = None,
    J_21_LW_grid: NDArray[float] | None = None,
    z_re_grid: NDArray[float] | None = None,
    Gamma12_grid: NDArray[float] | None = None,
):
    """
    Convert a halo catalogue's mass and RNG fields to halo properties.

    Assumes no feedback (Lyman-Werner, reionization).

    Returns a dict of 12 properties per halo:
        halo mass
        stellar mass (ACG)
        star formation rate (ACG)
        xray luminosity (combined)
        ionising emissivity (combined)
        escape-fraction weighted SFR (combined)
        stellar mass (MCG)
        star formation rate (MCG)
        ACG turnover mass
        MCG turnover mass
        Reionization turnover mass
        Metallicity
    """
    # single element zero array to act as the grids (vcb, J_21_LW, z_reion, Gamma12)
    if not (halo_masses.shape == star_rng.shape == sfr_rng.shape == xray_rng.shape):
        raise ValueError("Halo masses and rng shapes must be identical.")

    n_halos = halo_masses.size
    out_buffer = np.zeros((n_halos, 12), dtype="f4")
    lo_dim = (inputs.simulation_options.HII_DIM,) * 3

    if halo_coords is None:
        halo_coords = np.zeros(3 * n_halos)
    if vcb_grid is None:
        vcb_grid = np.zeros(lo_dim)
    if J_21_LW_grid is None:
        J_21_LW_grid = np.zeros(lo_dim)
    if z_re_grid is None:
        z_re_grid = np.zeros(lo_dim)
    if Gamma12_grid is None:
        Gamma12_grid = np.zeros(lo_dim)

    vcb_grid = vcb_grid.astype("f4")
    J_21_LW_grid = J_21_LW_grid.astype("f4")
    z_re_grid = z_re_grid.astype("f4")
    Gamma12_grid = Gamma12_grid.astype("f4")

    halo_masses = halo_masses.astype("f4")
    halo_coords = halo_coords.astype("f4")
    star_rng = star_rng.astype("f4")
    sfr_rng = sfr_rng.astype("f4")
    xray_rng = xray_rng.astype("f4")

    lib.test_halo_props(
        redshift,
        ffi.cast("float *", vcb_grid.ctypes.data),
        ffi.cast("float *", J_21_LW_grid.ctypes.data),
        ffi.cast("float *", z_re_grid.ctypes.data),
        ffi.cast("float *", Gamma12_grid.ctypes.data),
        n_halos,
        ffi.cast("float *", halo_masses.ctypes.data),
        ffi.cast("float *", halo_coords.ctypes.data),
        ffi.cast("float *", star_rng.ctypes.data),
        ffi.cast("float *", sfr_rng.ctypes.data),
        ffi.cast("float *", xray_rng.ctypes.data),
        ffi.cast("float *", out_buffer.ctypes.data),
    )

    out_buffer = out_buffer.reshape(n_halos, 12)

    return {
        "halo_mass": out_buffer[:, 0].reshape(halo_masses.shape),
        "halo_stars": out_buffer[:, 1].reshape(halo_masses.shape),
        "halo_sfr": out_buffer[:, 2].reshape(halo_masses.shape),
        "halo_xray": out_buffer[:, 3].reshape(halo_masses.shape),
        "n_ion": out_buffer[:, 4].reshape(halo_masses.shape),
        "halo_wsfr": out_buffer[:, 5].reshape(halo_masses.shape),
        "halo_stars_mini": out_buffer[:, 6].reshape(halo_masses.shape),
        "halo_sfr_mini": out_buffer[:, 7].reshape(halo_masses.shape),
        "mturn_a": out_buffer[:, 8].reshape(halo_masses.shape),
        "mturn_m": out_buffer[:, 9].reshape(halo_masses.shape),
        "mturn_r": out_buffer[:, 10].reshape(halo_masses.shape),
        "metallicity": out_buffer[:, 11].reshape(halo_masses.shape),
    }


@init_sigma_table
def return_uhmf_value(
    *,
    inputs: InputParameters,
    redshift: float,
    mass_values: Sequence[float],
):
    """Return the value of the unconditional halo mass function at given parameters.

    Parameters
    ----------
    inputs : InputParameters
        The input parameters defining the simulation run.
    redshift : float
        The redshift at which to evaluate the halo mass function.
    mass_values : float
        The mass values at which to evaluate the halo mass function.
    """
    growthf = lib.dicke(redshift)
    return np.vectorize(lib.unconditional_hmf)(
        growthf, np.log(mass_values), redshift, inputs.matter_options.cdict["HMF"]
    )


@init_sigma_table
def return_chmf_value(
    *,
    inputs: InputParameters,
    redshift: float,
    mass_values: Sequence[float],
    delta_values: Sequence[float],
    condmass_values: Sequence[float],
):
    """Return the value of the conditional halo mass function at given parameters.

    Parameters
    ----------
    inputs : InputParameters
        The input parameters defining the simulation run.
    redshift : float
        The redshift at which to evaluate the halo mass function.
    mass_values : float
        The mass values at which to evaluate the halo mass function.
    delta : float
        The overdensity at which to evaluate the halo mass function.
    cond_mass : float
        The condition mass at which to evaluate the halo mass function.
    """
    growthf = lib.dicke(redshift)
    sigma = np.vectorize(lib.sigma_z0)(condmass_values)

    return np.vectorize(lib.conditional_hmf)(
        growthf,
        np.log(mass_values[None, None, :]),
        delta_values[:, None, None],
        sigma[None, :, None],
        inputs.matter_options.cdict["HMF"],
    )



"""Functions for setting up and configuring inputs to driver functions."""

from __future__ import annotations

import contextlib
import functools
import inspect
import logging
from collections.abc import Sequence
from typing import Any, get_args

from .._cfg import config
from ..input_serialization import convert_inputs_to_dict
from ..io import h5
from ..io.caching import OutputCache
from ..utils import recursive_difference
from ..wrapper.cfuncs import (
    broadcast_input_struct,
    construct_fftw_wisdoms,
    free_cosmo_tables,
)
from ..wrapper.inputs import InputParameters
from ..wrapper.outputs import OutputStruct, OutputStructZ, _HashType
from ..wrapper.photoncons import _photoncons_state

logger = logging.getLogger(__name__)


def check_redshift_consistency(
    redshift: float, output_structs: list[OutputStruct], funcname: str = "unknown"
) -> None:
    """
    Check if all given :class:`OutputStruct` objects have the same redshift.

    This function iterates over a list of OutputStruct objects and verifies that all
    the redshifts are consistent with the provided redshift value. If any inconsistency
    is found, a ValueError is raised.

    Parameters
    ----------
    redshift
        The reference redshift value to compare with the redshifts of the OutputStruct
        objects.
    output_structs
        A list of :class:`OutputStruct` objects to check for redshift consistency.
    funcname (str, optional)
        The name of the function or method where this check is being performed. Used
        in the error raised. Default is "unknown".

    Raises
    ------
    ValueError
        If any :class:`OutputStruct` object in the list has a redshift different from
        the provided redshift value.
    """
    for struct in output_structs:
        if struct is not None and struct.redshift != redshift:
            raise ValueError(
                f"Incompatible redshifts with inputs and {struct.__class__.__name__} in"
                f" {funcname}: {redshift} != {struct.redshift}"
            )


def _get_incompatible_params(
    inputs1: InputParameters, inputs2: InputParameters
) -> dict[str, Any]:
    """Return a dict of parameters that differ between two InputParameters objects."""
    d1 = convert_inputs_to_dict(inputs1, only_structs=False, camel=False)
    d2 = convert_inputs_to_dict(inputs2, only_structs=False, camel=False)
    return recursive_difference(d1, d2)


def _get_incompatible_param_diffstring(
    inputs1: InputParameters, inputs2: InputParameters
) -> str:
    incompatible_params = _get_incompatible_params(inputs1, inputs2)
    rev = _get_incompatible_params(inputs2, inputs1)

    return "".join(
        (
            f"{name}:\n"
            + "\n".join(
                f"  {key}:\n    {v1:>12}\n    {rev[name][key]:>12}"
                for key, v1 in val.items()
            )
            if isinstance(val, dict)
            else f"{name}:\n  {val}\n  {rev[name]}"
        )
        for name, val in incompatible_params.items()
    )


def check_output_consistency(outputs: dict[str, OutputStruct]) -> None:
    """Ensure all OutputStruct objects have consistent InputParameters.

    This function compares each given OutputStruct with a reference element, and ensures
    that each is compatible. Recall that two :class:`InputParameters` can be compatible
    even if they differ, as long as they agree on the input components to which each
    of the OutputStructs are dependent.

    Raises
    ------
    ValueError
        If any of the OutputStructs are not compatible.
    """
    outputs = {n: output for n, output in outputs.items() if output is not None}

    if len(outputs) < 2:
        return

    o0 = next(iter(outputs.values()))
    n0 = next(iter(outputs.keys()))

    for name, output in outputs.items():
        if not output._inputs_compatible_with(o0):
            diff = _get_incompatible_param_diffstring(output.inputs, o0.inputs)
            raise ValueError(
                f"InputParameters in {name} do not match those in {n0}. Got:\n\n{diff}"
            )


def check_consistency_of_outputs_with_inputs(
    inputs: InputParameters, outputs: Sequence[OutputStruct]
):
    """Check that all structs in `outputs` are compatible with the `inputs`.

    See Also
    --------
    :func:`check_output_consistency`
        Similar function that checks consistency between several outputs.
    """
    for output in outputs:
        if not output._inputs_compatible_with(inputs):
            diff = _get_incompatible_param_diffstring(output.inputs, inputs)
            raise ValueError(
                f"InputParameters in {output.__class__.__name__} do not match those in "
                f"the provided InputParameters. Got:\n\n{diff}"
            )


class _OutputStructComputationInspect:
    """A class that does introspection on a single-field computation function.

    This class implements methods for inspecting the arguments of a single-field
    computation function (e.g. :func:`compute_initial_conditions`) and doing validation,
    cache-checking and other quality-of-life improvements.

    It is an internal toolset, not meant to be used by users directly.
    """

    def __init__(self, _func: callable):
        self._func = _func
        self._signature = inspect.signature(_func)
        self._kls = self._signature.return_annotation

        if not issubclass(self._signature.return_annotation, OutputStruct):
            raise TypeError(
                f"{_func.__name__} must return an instance of OutputStruct (and be annotated as such)."
            )

    @staticmethod
    def _get_all_output_struct_inputs(
        kwargs, recurse: bool = False
    ) -> dict[str, OutputStruct]:
        """Return all the arguments that are OutputStructs.

        If recurse is True, also add all OutputStructs that are part of iterables.
        """
        d = {k: v for k, v in kwargs.items() if isinstance(v, OutputStruct)}

        if recurse:
            for k, v in kwargs.items():
                if hasattr(v, "__len__") and isinstance(v[0], OutputStruct):
                    d |= {f"{k}_{i}": vv for i, vv in enumerate(v)}

        return d

    @staticmethod
    def _get_inputs(kwargs: dict[str, Any]) -> InputParameters:
        """Return the most detailed input parameters available.

        For a given set of parameters to a single-field function, find the "inputs"
        that should be used for instantiating the OutputStruct that the function will
        return.

        If the parameter "inputs" is given directly, just return that. If not, the
        inputs must be determined from given OutputStruct parameters that are
        dependencies of the desired OutputStruct. In this case, we can run into the
        situation that different dependent OutputStruct's have different inputs. Even
        though all must be compatible with each other, more basic OutputStructs (like
        InitialConditions) might not have the same zgrid as the PerturbedField (for
        example) and this is fine. So, here we return the inputs of the "most advanced"
        OutputStruct that is given.
        """
        inputs = kwargs.get("inputs")
        if inputs is not None:
            return inputs

        outputs = _OutputStructComputationInspect._get_all_output_struct_inputs(
            kwargs, recurse=True
        )

        minreq = _HashType(0)
        for output in outputs.values():
            if output._compat_hash.value >= minreq.value:
                inputs = output.inputs
                minreq = output._compat_hash

        if inputs is None:
            raise ValueError(
                "No parameter 'inputs' given, and no dependent OutputStruct found!"
            )

        return inputs

    @staticmethod
    def check_consistency(kwargs: dict[str, Any], outputs: dict[str, OutputStruct]):
        """Check consistency of input parameters amongst output struct inputs."""
        check_output_consistency(outputs)
        given_inputs = kwargs.get("inputs")
        if given_inputs is not None:
            check_consistency_of_outputs_with_inputs(given_inputs, outputs.values())

    def _make_wisdoms(self, use_fftw_wisdom: bool):
        construct_fftw_wisdoms(use_fftw_wisdom=use_fftw_wisdom)

    def _broadcast_inputs(self, inputs: InputParameters):
        broadcast_input_struct(inputs=inputs)

    def _free_cosmo_tables(self):
        free_cosmo_tables()

    def check_output_struct_types(self, outputs: dict[str, OutputStruct]):
        """Check given OutputStruct parameters.

        This method checks each OutputStruct given to the compuation function to ensure
        that it is of the correct type (i.e. `InitialConditions` when requested). The
        types must be specified as standard Python types (so this is essentially just
        automated run-time type-checking). It only checks OutputStructs, and allows for
        optional values.
        """
        for name, param in self._signature.parameters.items():
            val = outputs.get(name)
            tp = param.annotation
            try:
                issub = issubclass(tp, OutputStruct)
            except TypeError:
                # The parameter type is not a subclass of OutputStruct, ignore it.
                issub = False
            if issub and not isinstance(val, tp):
                raise TypeError(
                    f"{name} should be of type {param.annotation.__name__}, got {type(val)}"
                )

            if potential_types := get_args(tp):
                if type(None) in potential_types and val is None:
                    continue
                kls = tuple(
                    kls for kls in potential_types if issubclass(kls, OutputStruct)
                )
                if not kls:
                    # This is not an OutputStruct kind of parameter, ignore.
                    continue
                elif len(kls) > 1:
                    raise TypeError(
                        f"{name} parameter has a badly defined type in the signature. Please report this on Github."
                    )
                else:
                    if type(None) not in potential_types:
                        # This is supposed to be a list/sequence of OutputStruct,
                        # but this is hard to check because these values are already
                        # stripped out of the `outputs` variable that is passed here.
                        # So for now, just ignore it.
                        continue

                    kls = kls[0]

                    if not isinstance(val, kls):
                        raise TypeError(
                            f"{name} should be of type {kls.__name__}, got {type(val)}"
                        )

    def _get_current_redshift(
        self, outputs: dict[str, OutputStructZ], kwargs: dict[str, Any]
    ) -> float | None:
        """Get the current redshift of evolution from the given parameters.

        If redshift is given directly, return that. Otherwise, return a redshift from
        any given OutputStruct whose name doesn't start with "previous" or "descendant".
        We can return the first such redshift we find, because another method will check
        that all redshifts are the same.
        """
        redshift = kwargs.get("redshift")
        if redshift is None and (
            current_outputs := [
                v
                for k, v in outputs.items()
                if not k.startswith("previous_") and not k.startswith("descendant_")
            ]
        ):
            redshift = current_outputs[0].redshift

        return redshift

    def ensure_redshift_consistency(
        self, current_redshift: float, outputs: dict[str, OutputStructZ]
    ):
        """Ensure that each OutputStruct has the same redshift, if it exists.

        Checks all given OutputStruct objects that have redshifts to check if their
        redshifts are the same. It does this in three groups: previous, descendant and
        current-redshift boxes.
        """
        if not outputs:
            return

        if current_outputs := [
            v
            for k, v in outputs.items()
            if not k.startswith("previous_") and not k.startswith("descendant_")
        ]:
            check_redshift_consistency(
                current_redshift, current_outputs, funcname=self._func.__name__
            )

        inputs = next(iter(outputs.values())).inputs
        if inputs.node_redshifts is not None:
            previous_outputs = [
                v for k, v in outputs.items() if k.startswith("previous_")
            ]
            previous_z = [z for z in inputs.node_redshifts if z > current_redshift]
            if previous_outputs and previous_z:
                previous_z = previous_z[-1]
                check_redshift_consistency(
                    previous_z,
                    previous_outputs,
                    funcname=f"{self._func.__name__} (previous z)",
                )

            descendant_outputs = [
                v for k, v in outputs.items() if k.startswith("descendant_")
            ]
            descendant_z = [z for z in inputs.node_redshifts if z < current_redshift]

            if descendant_outputs and descendant_z:
                descendant_z = descendant_z[0]
                check_redshift_consistency(
                    descendant_z,
                    descendant_outputs,
                    funcname=f"{self._func.__name__} (descendant z)",
                )

    def check_backend_state(self, inputs: InputParameters):
        """Check the backend state of the computation function.

        Currently only holds the check for whether the photon conservation is
        both needed and not initialised.
        In Future, may hold more backend state checks.
        """
        # we need photon cons to be done before any non-IC box is computed
        if (
            inputs.astro_options.PHOTON_CONS_TYPE != "no-photoncons"
            and _photoncons_state.calibration_inputs != inputs
            and issubclass(self._kls, OutputStructZ)
        ):
            raise ValueError(
                "Photon conservation is needed but not initialised with the current InputParameters."
                " Call `setup_photon_cons` with your current parameters or use the high-level functions."
            )

    def _handle_read_from_cache(
        self,
        inputs: InputParameters,
        current_redshift: float | None,
        cache: OutputCache | None,
        regen: bool = True,
    ) -> OutputStruct | None:
        """Handle potential reading from cache.

        Checks the given input parameters for cache-related keywords and manages reading
        an OutputStruct from cache if possible and desired.
        """
        if cache is None or regen:
            return None

        # First check whether the boxes already exist.
        if issubclass(self._kls, OutputStructZ):
            obj = self._kls.new(inputs=inputs, redshift=current_redshift)
        else:
            obj = self._kls.new(inputs=inputs)

        path = cache.find_existing(obj)
        if path is not None:
            with contextlib.suppress(OSError):
                this = h5.read_output_struct(path, safe=config["safe_read"])
                if hasattr(this, "redshift"):
                    logger.info(
                        f"Existing {obj._name} found at z={this.redshift} and read in (seed={this.random_seed})."
                    )
                else:
                    logger.info(
                        f"Existing {obj._name} found and read in (seed={this.random_seed})."
                    )
                return this

    def _handle_write_to_cache(
        self, cache: OutputCache | None, write, obj: OutputStruct
    ):
        """Handle writing a box to cache."""
        if write and not cache:
            raise ValueError("Cannot write to cache without a cache object.")

        if write:
            cache.write(obj)


class single_field_func(_OutputStructComputationInspect):  # noqa: N801
    """A decorator for functions that compute single fields.

    This decorator is meant for internal use only.
    """

    def __init__(self, _func: callable):
        super().__init__(_func)

        # Explicitly update the wrapper instance with the original function's metadata
        functools.update_wrapper(self, _func)

    def __call__(self, **kwargs) -> OutputStruct:
        """Call the single field function."""
        inputs = self._get_inputs(kwargs)
        outputs = self._get_all_output_struct_inputs(kwargs)
        outputs_rec = self._get_all_output_struct_inputs(kwargs, recurse=True)
        outputsz = {k: v for k, v in outputs.items() if isinstance(v, OutputStructZ)}

        # Get current redshift (could be None)
        current_redshift = self._get_current_redshift(outputsz, kwargs)

        self.check_consistency(kwargs, outputs_rec)
        self.check_output_struct_types(outputs)
        # The following checks both current and previous redshifts, if applicable
        self.ensure_redshift_consistency(current_redshift, outputsz)
        self.check_backend_state(inputs)

        cache = kwargs.pop("cache", None)
        regen = kwargs.pop("regenerate", True)
        write = kwargs.pop("write", False)

        out = self._handle_read_from_cache(inputs, current_redshift, cache, regen)

        free_cosmo_tables = kwargs.pop("free_cosmo_tables", True)

        if "inputs" in self._signature.parameters:
            # Here we set the inputs (if accepted by the function signature)
            # to the most advanced ones. This is the explicitly-passed inputs if
            # they exist, but otherwise the inputs derived from the dependency
            # that is the most advanced in the computation.
            kwargs["inputs"] = inputs

        if out is None:
            self._broadcast_inputs(inputs)
            self._make_wisdoms(inputs.matter_options.USE_FFTW_WISDOM)
            out = self._func(**kwargs)
            self._handle_write_to_cache(cache, write, out)

        # Free cosmo_tables, unless it is explicitly requested to keep their memory
        # (useful in macro functions like run_coeval and run_lightcone, so we won't have to
        # free and reallocate memory repeatedly)
        if free_cosmo_tables:
            self._free_cosmo_tables()

        return out


class high_level_func(_OutputStructComputationInspect):  # noqa: N801
    """A decorator for high-level functions like ``run_coeval``."""

    def __init__(self, _func: callable):
        self._func = _func
        self._signature = inspect.signature(_func)
        self._kls = self._signature.return_annotation
        functools.update_wrapper(self, _func)

    def __call__(self, **kwargs):
        """Call the function."""
        outputs = self._get_all_output_struct_inputs(kwargs, recurse=True)
        inputs = self._get_inputs(kwargs)
        if "inputs" in self._signature.parameters:
            # Here we set the inputs (if accepted by the function signature)
            # to the most advanced ones. This is the explicitly-passed inputs if
            # they exist, but otherwise the inputs derived from the dependency
            # that is the most advanced in the computation.
            kwargs["inputs"] = inputs

        self.check_consistency(kwargs, outputs)

        yield from self._func(**kwargs)



"""Configuration options for the docs."""

import os

# import sys
# from pathlib import Path
# from unittest.mock import MagicMock
# Import the package because if the import breaks, this gives a nice
# stack trace, whereas if it breaks inside the autodoc step, it's harder to debug.
import py21cmfast

extensions = [
    "sphinx.ext.autodoc.typehints",
#    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.ifconfig",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.mathjax",
    "sphinx.ext.autosectionlabel",
    "autoapi.extension",
    "numpydoc",
    "nbsphinx",
    "sphinx_design",
    "sphinx_copybutton",
    "IPython.sphinxext.ipython_console_highlighting",
]
if os.getenv("SPELLCHECK"):
    extensions += ("sphinxcontrib.spelling",)
    spelling_show_suggestions = True
    spelling_lang = "en_US"

autosectionlabel_prefix_document = True
# autodoc_use_legacy_class_based = True

# Auto-API settings
autoapi_options=[
    "members",
    "undoc-members",
    "show-inheritance",
    "show-module-summary",
    "special-members",
    "imported-members",
    "inherited-members",
]
autoapi_dirs = ["../src/py21cmfast"]
autoapi_add_toctree_entry = False  # We add it ourselves in index.rst
autoapi_python_class_content = "init"
autoapi_member_order = "groupwise"
autoapi_own_page_level = "class"
autoapi_keep_files = True
autodoc_typehints = "description"

autosummary_generate = False
numpydoc_show_class_members = False

source_suffix = ".rst"
master_doc = "index"
project = "21cmFAST"
year = "2020"
author = "The 21cmFAST collaboration"
copyright = f"{year}, {author}"
version = release = py21cmfast.__version__
templates_path = ["templates"]

pygments_style = "trac"
extlinks = {
    "issue": ("https://github.com/21cmFAST/21cmFAST/issues/%s", "#"),
    "pr": ("https://github.com/21cmFAST/21cmFAST/pull/%s", "PR #"),
}
# # on_rtd is whether we are on readthedocs.org
# on_rtd = os.environ.get("READTHEDOCS", None) == "True"

html_theme = "furo"
html_style = "css/custom.css"
html_logo = "images/Logo_square_transparent.png"

# These folders are copied to the documentation's HTML output
html_static_path = ["_static"]
html_css_files = [
    "css/custom.css",
]

html_theme_options = {
    "sidebar_hide_name": True,
}
html_use_smartypants = True
html_last_updated_fmt = "%b %d, %Y"
html_split_index = False
html_sidebars = {
    "**": [
        "sidebar/scroll-start.html",
        "sidebar/brand.html",
        "sidebar/search.html",
        "sidebar/navigation.html",
        "sidebar/ethical-ads.html",
        "sidebar/scroll-end.html",
    ]
}
html_short_title = f"{project}-{version}"

napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_use_param = False

mathjax_path = (
    "http://cdn.mathjax.org/mathjax/latest/MathJax.js?config=TeX-AMS-MML_HTMLorMML"
)

exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "templates",
    "**.ipynb_checkpoints",
]



"""Simple script to run a lightcone for performance testing."""

import sys
import time
from pathlib import Path

import tyro

from py21cmfast import OutputCache, RectilinearLightconer, run_lightcone

here = Path(__file__).parent.resolve()

sys.path.append(str(here / "tests"))

from produce_integration_test_data import (  # noqa: E402
    OPTIONS_TESTRUNS,
    get_all_options_struct,
    get_lc_fields,
)


def get_at_it(
    case: str = "fixed_halogrids",
    sizefac: int = 1,
):
    """Run the lightcone."""
    t = time.time()

    z, kwargs = OPTIONS_TESTRUNS[case]
    inputs = get_all_options_struct(z, lc=True, **kwargs)["inputs"]

    inputs = inputs.evolve_input_structs(
        HII_DIM=inputs.simulation_options.HII_DIM * sizefac,
        BOX_LEN=inputs.simulation_options.BOX_LEN * sizefac,
        DIM=inputs.simulation_options.DIM * sizefac,
    )

    node_z = inputs.node_redshifts

    quantities = get_lc_fields(inputs)
    try:
        lcn = RectilinearLightconer.between_redshifts(
            min_redshift=node_z[-1] + 0.2,
            max_redshift=node_z[0] - 0.2,
            quantities=quantities,
            resolution=inputs.simulation_options.cell_size,
        )
    except AttributeError:
        lcn = RectilinearLightconer.with_equal_cdist_slices(
            min_redshift=node_z[-1] + 0.2,
            max_redshift=node_z[0] - 0.2,
            quantities=quantities,
            resolution=inputs.simulation_options.cell_size,
        )
    run_lightcone(
        lightconer=lcn,
        write=True,  # write so that perturbed fields and halos can be cached.
        regenerate=True,
        cache=OutputCache("."),
        inputs=inputs,
        progressbar=True,
    )
    print(f"Lightcone run took {time.time() - t:.2f} seconds.")


if __name__ == "__main__":
    tyro.cli(get_at_it, description="Run a lightcone for performance testing.")
    # Example usage: python -m tests.performance --case fixed_halogrids --sizefac 2
    # This will run the lightcone with double the resolution and box size of the fixed_halogrids case.



"""Functions for handling InputParameters for various purposes, e.g. I/O and unstructuring."""

import warnings
from typing import Any, Literal

import attrs

from .utils import recursive_difference
from .wrapper._utils import camel_to_snake, snake_to_camel
from .wrapper.inputs import CosmoTables, InputParameters, InputStruct, Table1D


def convert_inputs_to_dict(
    inputs: InputParameters,
    mode: Literal["full", "minimal"] = "full",
    only_structs: bool = True,
    camel: bool = True,
    only_cstruct_params: bool = False,
    use_aliases: bool = True,
) -> dict[str, dict[str, Any]]:
    """Convert an InputParameters object to a dictionary, with various options.

    While the :class:`~py21cmfast.wrapper.inputs.InputParameters.asdict` method allows
    several options for "unstructuring" the inputs to a dictionary, this function is
    much more flexible, having options that cover a number of use-cases.

    Parameters
    ----------
    inputs
        The input parameters to convert to a dict.
    mode
        Either 'minimal' to get the minimal dictionary required to specify the
        parameters (on top of defaults) or 'full' to keep all parameters.
    only_structs
        Whether to only return InputStruct objects (unstructured to dicts),
        otherwise also return other attributes (e.g. random_seed).
    camel
        Whether the keys of the returned dict should be camel-case, e.g.
        SimulationOptions. Otherwise, return as snake_case. Only applies to
        InputStruct attributes of the inputs (i.e. not node_redshifts or random_seed).
    only_cstruct_params
        Only return parameters that are part of the Cstruct, rather than all
        fields of the class. This is useful for pretty-printing.
    use_aliases
        If True, use correct aliases for parameters, which allows the dictionary
        to be passed back into the InputStruct constructors (e.g. use DIM instead
        of _DIM).
    """
    kw = {
        "only_structs": only_structs,
        "camel": camel,
        "only_cstruct_params": only_cstruct_params,
        "use_aliases": use_aliases,
    }
    all_inputs = inputs.asdict(**kw)

    if mode == "minimal":
        defaults = InputParameters(random_seed=0)
        default_dct = defaults.asdict(**kw)
        cosmo_tables_key = "CosmoTables" if camel else "cosmo_tables"
        # we still want to keep cosmo_tables, even in minimal mode (since we want to keep ps_norm and USE_SIGMA_8)
        cosmo_tables_dct = all_inputs[cosmo_tables_key].copy()
        all_inputs = recursive_difference(all_inputs, default_dct)
        all_inputs[cosmo_tables_key] = cosmo_tables_dct

    return all_inputs


def prepare_inputs_for_serialization(
    inputs: InputParameters,
    mode: Literal["full", "minimal"] = "full",
    only_structs: bool = True,
    camel: bool = True,
) -> dict[str, dict[str, Any]]:
    """Prepare an inputs class for serialization (to e.g. TOML, YAML or HDF5).

    This is a thin wrapper around :func:`~convert_inputs_to_dict` that also
    ensures that 'None' values are removed (so long as their default is also None)
    and that the parameter names map back to aliases of InputStruct attributes.
    """
    dct = convert_inputs_to_dict(
        inputs,
        mode=mode,
        only_structs=only_structs,
        camel=camel,
        use_aliases=False,  # convert to aliases below instead.
    )

    # dct is a dict of dicts, where each subdict represents a
    # single input parameter struct (SimulationOptions, AstroParams etc).
    # The rules are that _all_ of the input structs are in this dict, even if they are
    # empty. The keys allowed in each struct are such that:
    #
    #    1. They are valid keys to pass to the class constructor, not necessarily
    #       all the fields required for the Cstruct (since sometimes these are
    #       auto-calcualted).
    #    2. They are not necessarily exhaustive, especially if mode=="minimal".
    #    3. The values are those that are required to *build the isntance*, not their
    #       final transformed values for use in C.
    #
    # However, some attributes of the InputStruct classes are not necessarily
    # required for building the Cstruct. This can occur if there are two pathways of
    # setting a particular necessary parameter (e.g. DIM can be set either by setting
    # DIM directly, or by setting HII_DIM and HIRES_TO_LOWRES_FACTOR). In this case,
    # at least one of the attributes will have a default value of None. If something
    # has an actual value of None, we require that its default is None. This way,
    # we can simply leave it out of the written TOML (or HDF5) when its value is None,
    # and it will anyway be set to its own default if read back in.
    out = {}
    for structname, structvals in dct.items():
        this = {}
        clsname = snake_to_camel(structname)
        if clsname in InputStruct._subclasses:
            fields = attrs.fields_dict(InputStruct._subclasses[clsname])
        elif clsname == "CosmoTables":
            fields = attrs.fields_dict(CosmoTables)

        for key, val in structvals.items():
            if val is None:
                if fields[key].default is not None:  # pragma: nocover
                    # This should not be reachable because setting a required parameter
                    # to None should error on validation, rather than reaching here.
                    raise RuntimeError(
                        f"Detected that {structname} has {key}=None but it is not an optional parameter!"
                    )
            else:
                # Convert to use the alias, rather than the name (i.e. DIM instead of _DIM)
                this[fields[key].alias] = val
        if this:
            out[structname] = this

    # Furthermore, some
    return out


def deserialize_inputs(
    dict_of_structdicts: dict[str, Any], safe: bool = True, **loose_params
) -> dict[str, InputStruct]:
    """Construct a dictionary of InputStructs ready to be converted to InputParameters.

    Parameters
    ----------
    dict_of_structdicts
        A dictionary whose keys are names of InputStruct attributes of
        InputParameters (e.g. CosmoParams, SimulationOptions), and whose values
        are dictionaries of parameters specific to each struct. Not every parameter
        of every struct is required.

    Other Parameters
    ----------------
    All other parameters are considered "loose" parameters of one of the InputStructs,
    and will *override* the parameter matching their name if it is found in one of the
    structs.

    Returns
    -------
    dict_of_structs
        A dictionary whose keys are attribute names of the InputParameters class,
        and whose values are InputStruct instances, ready to instantiate an
        InputParameters class, e.g. ``InputParameters(**dict_of_structs, random_seed=1)``.
    """
    # It's possible that names in dict_of_structdicts are in camel case or snake
    # case. We normalise to camel case here. This also de-references the input
    # so when we modify it in-place (via .pop()) later, we don't mess with the user's
    # input.
    dict_of_structdicts = {
        snake_to_camel(name): dct for name, dct in dict_of_structdicts.items()
    }

    input_dict = {}
    extra_params = {}
    for structname, kls in (
        InputStruct._subclasses | {"CosmoTables": CosmoTables}
    ).items():
        # Use field.alias instead of field.name because the alias is what needs
        # to be passed to the class constructor (e.g. "DIM" instead of "_DIM")
        fieldnames = [field.alias for field in attrs.fields(kls)]
        kw_dict = {kk: loose_params.pop(kk) for kk in fieldnames if kk in loose_params}

        these_all = dict_of_structdicts.pop(structname, {})

        if structname == "CosmoTables":
            these = {
                kk: Table1D(**these_all[kk])
                if isinstance(these_all[kk], dict)
                else these_all[kk]
                for kk in these_all
                if kk in fieldnames
            }
        else:
            these = {kk: these_all[kk] for kk in these_all if kk in fieldnames}
        extra = {kk: these_all[kk] for kk in these_all if kk not in fieldnames}

        # Here, if structname is not in the input dict_of_structdicts, it is OK,
        # because we just assume an empty set of parameters, potentially added to
        # by the loose params.
        arg_dict = {**these, **kw_dict}
        input_struct = kls.new(arg_dict)
        input_dict[camel_to_snake(structname)] = input_struct

        if extra:
            extra_params[structname] = extra

    if dict_of_structdicts:
        warnings.warn(
            f"The following keys were not recognized for deserializing to InputParameters: {tuple(dict_of_structdicts.keys())}",
            stacklevel=2,
        )

    if extra_params or loose_params:
        all_extra = {**extra_params, **loose_params}
        msg = f"Excess arguments exist: {all_extra}"
        if safe:
            raise ValueError(msg)
        else:
            warnings.warn(msg, stacklevel=2)

    return input_dict



"""Simple example script for plotting estimated memory usage."""

import matplotlib.pyplot as plt
import numpy as np
from py21cmfast._memory import estimate_memory_lightcone

from py21cmfast import AstroFlags, AstroParams, CosmoParams, MatterParams

matter_params = MatterParams(
    {
        "HII_DIM": 250,
        "BOX_LEN": 2000.0,
        "DIM": 1000,
        "N_THREADS": 8,
        "USE_RELATIVE_VELOCITIES": True,
        "POWER_SPECTRUM": 5,
        "USE_FFTW_WISDOM": True,
        "PERTURB_ON_HIGH_RES": True,
        "USE_INTERPOLATION_TABLES": True,
        "FAST_FCOLL_TABLES": True,
    }
)
astro_flags = AstroFlags(
    {
        "INHOMO_RECO": True,
        "USE_TS_FLUCT": True,
        "USE_MINI_HALOS": True,
    },
    USE_VELS_AUX=True,
)


h2dim = np.array(list(range(200, 1500, 100)))


mems4 = []
for h2 in range(200, 1500, 100):
    matter_params.update(HII_DIM=h2, DIM=4 * h2)
    mem = estimate_memory_lightcone(
        max_redshift=1420 / 50 - 1,
        redshift=1420 / 200 - 1,
        matter_params=matter_params,
        astro_flags=astro_flags,
        cosmo_params=CosmoParams(),
        astro_params=AstroParams(),
    )
    mems4.append(mem)
peaks4 = np.array([m["peak_memory"] / (1024**3) for m in mems4])

mems3 = []
for h2 in range(200, 1500, 100):
    matter_params.update(HII_DIM=h2, DIM=3 * h2)
    mem = estimate_memory_lightcone(
        max_redshift=1420 / 50 - 1,
        redshift=1420 / 200 - 1,
        matter_params=matter_params,
        astro_flags=astro_flags,
        cosmo_params=CosmoParams(),
        astro_params=AstroParams(),
    )
    mems3.append(mem)
peaks3 = np.array([m["peak_memory"] / (1024**3) for m in mems3])

mems2 = []
for h2 in range(200, 1500, 100):
    matter_params.update(HII_DIM=h2, DIM=2 * h2)
    mem = estimate_memory_lightcone(
        max_redshift=1420 / 50 - 1,
        redshift=1420 / 200 - 1,
        matter_params=matter_params,
        astro_flags=astro_flags,
        cosmo_params=CosmoParams(),
        astro_params=AstroParams(),
    )
    mems2.append(mem)
peaks2 = np.array([m["peak_memory"] / (1024**3) for m in mems2])


matter_params.update(HII_DIM=1300, DIM=2600)  # what we actually did.
mem = estimate_memory_lightcone(
    max_redshift=1420 / 50 - 1,
    redshift=1420 / 200 - 1,
    matter_params=matter_params,
    astro_flags=astro_flags,
    cosmo_params=CosmoParams(),
    astro_params=AstroParams(),
)


plt.plot(h2dim, peaks2, label="DIM=2*HII_DIM")
plt.plot(h2dim, peaks3, label="DIM=3*HII_DIM")
plt.plot(h2dim, peaks4, label="DIM=4*HII_DIM")

plt.scatter(
    1300,
    mem["peak_memory"] / 1024**3,
    marker="*",
    color="k",
    label="Proposed Sims",
    s=100,
    zorder=3,
)

plt.scatter(
    1300, 2924.847, marker="+", color="r", label="Measured Peak RAM", s=75, zorder=3
)

theory = h2dim**3
plt.plot(
    h2dim,
    theory * (peaks3.max() / theory.max()) * 1.2,
    label="HII_DIM^3 scaling",
    color="grey",
    ls="--",
)

plt.axhline(4000, linestyle="dashed", color="k")
plt.text(h2dim[0], 4300, "Avail RAM on Bridges2-EM")
plt.legend()
plt.yscale("log")
plt.xscale("log")
plt.ylabel("Peak memory (GB)")
plt.xlabel("HII_DIM")

plt.savefig("peak_memory_usage.png")



"""Unit tests for output structures."""

import pickle

import attrs
import numpy as np
import pytest

from py21cmfast import (
    InitialConditions,  # An example of an output struct
    InputParameters,
    config,
)
from py21cmfast.wrapper import outputs as ox
from py21cmfast.wrapper.arrays import Array


@pytest.fixture
def init(default_input_struct: InputParameters):
    return InitialConditions.new(inputs=default_input_struct)


def test_different_seeds(
    init: InitialConditions,
    default_input_struct: InputParameters,
):
    ic2 = InitialConditions.new(
        inputs=default_input_struct.clone(
            random_seed=default_input_struct.random_seed + 1
        )
    )

    assert init is not ic2
    assert init != ic2

    # make sure we didn't inadvertantly set the random seed while doing any of this
    assert init.random_seed == default_input_struct.random_seed


def test_pickleability(default_input_struct: InputParameters):
    ic_ = InitialConditions.new(inputs=default_input_struct)
    s = pickle.dumps(ic_)

    ic2 = pickle.loads(s)
    assert repr(ic_) == repr(ic2)


def test_reading_purged(ic: InitialConditions):
    lowres_density = ic.get(ic.lowres_density)

    # Remove it from memory
    ic.purge()

    assert not ic.lowres_density.state.computed_in_mem
    assert ic.lowres_density.state.on_disk

    # But we can still get it.
    lowres_density_2 = ic.get(ic.lowres_density)

    assert ic.lowres_density.state.on_disk
    assert ic.lowres_density.state.computed_in_mem

    assert np.allclose(lowres_density_2, lowres_density)

    ic.load_all()


@pytest.mark.parametrize("struct", list(ox._ALL_OUTPUT_STRUCTS.values()))
def test_all_fields_exist(struct: ox.OutputStruct):
    cstruct = ox.StructWrapper(struct.__name__)

    this = attrs.fields_dict(struct)

    # Ensure that all fields in the cstruct are also defined on this class.
    for name in cstruct.pointer_fields:
        assert name in this
        assert this[name].type == ox.Array

    for name in cstruct.primitive_fields:
        assert name in this


# NOTE: These do not test every field, but does test every conditional in the
#   OutputStruct constructors, a better approach would probably be to have a
#   comprehensive list of {"field_name": {"flag": value}} conditions for the fields
#   in the output module which is checked in the constructors
def test_optional_field_ic(default_input_struct_lc: InputParameters):
    """Ensure that the correct InitialConditions fields are set based on the parameters."""
    ic = ox.InitialConditions.new(inputs=default_input_struct_lc)
    assert isinstance(ic.lowres_vx, Array)
    assert isinstance(ic.lowres_vx_2LPT, Array)
    assert ic.hires_vx is None
    assert isinstance(ic.hires_vx_2LPT, Array)  # Python requires it, check the C
    assert ic.lowres_vcb is None

    ic = ox.InitialConditions.new(
        inputs=default_input_struct_lc.evolve_input_structs(
            PERTURB_ALGORITHM="ZELDOVICH"
        )
    )
    assert isinstance(ic.lowres_vy, Array)
    assert ic.lowres_vy_2LPT is None
    assert ic.hires_vy is None
    assert ic.hires_vy_2LPT is None

    ic = ox.InitialConditions.new(
        inputs=default_input_struct_lc.evolve_input_structs(PERTURB_ON_HIGH_RES=True)
    )
    assert ic.lowres_vz is None
    assert ic.lowres_vz_2LPT is None
    assert isinstance(ic.hires_vz, Array)
    assert isinstance(ic.hires_vz_2LPT, Array)

    ic = ox.InitialConditions.new(
        inputs=default_input_struct_lc.evolve_input_structs(
            USE_RELATIVE_VELOCITIES=True,
            POWER_SPECTRUM="CLASS",
        )
    )
    assert isinstance(ic.lowres_vx, Array)
    assert isinstance(ic.lowres_vx_2LPT, Array)
    assert ic.hires_vx is None
    assert isinstance(ic.hires_vx_2LPT, Array)
    assert isinstance(ic.lowres_vcb, Array)


def test_optional_field_perturb(default_input_struct_lc: InputParameters):
    """Ensure that the correct PerturbedField fields are set based on the parameters."""
    pt = ox.PerturbedField.new(redshift=0.0, inputs=default_input_struct_lc)
    assert isinstance(pt.density, Array)
    assert isinstance(pt.velocity_z, Array)
    assert isinstance(pt.velocity_x, Array)
    assert isinstance(pt.velocity_y, Array)

    pt = ox.PerturbedField.new(
        redshift=0.0,
        inputs=default_input_struct_lc.evolve_input_structs(KEEP_3D_VELOCITIES=False),
    )
    assert isinstance(pt.density, Array)
    assert isinstance(pt.velocity_z, Array)
    assert pt.velocity_x is None
    assert pt.velocity_y is None


def test_optional_field_perturbed_halocat(default_input_struct_lc: InputParameters):
    """Ensure that the correct HaloBox fields are set based on the parameters."""
    pert_halo_cat = ox.PerturbedHaloCatalog.new(
        redshift=0.0, inputs=default_input_struct_lc, buffer_size=1
    )
    assert isinstance(pert_halo_cat.halo_masses, Array)
    assert isinstance(pert_halo_cat.halo_coords, Array)
    assert isinstance(pert_halo_cat.halo_masses, Array)
    assert isinstance(pert_halo_cat.halo_coords, Array)
    assert isinstance(pert_halo_cat.stellar_masses, Array)
    assert isinstance(pert_halo_cat.ion_emissivity, Array)
    assert pert_halo_cat.xray_emissivity is None
    assert pert_halo_cat.fesc_sfr is None
    assert pert_halo_cat.stellar_mini is None
    assert pert_halo_cat.sfr_mini is None

    inputs = default_input_struct_lc.evolve_input_structs(USE_TS_FLUCT=True)
    pert_halo_cat = ox.PerturbedHaloCatalog.new(
        redshift=0.0, inputs=inputs, buffer_size=1
    )
    assert isinstance(pert_halo_cat.xray_emissivity, Array)
    inputs = inputs.evolve_input_structs(INHOMO_RECO=True)
    pert_halo_cat = ox.PerturbedHaloCatalog.new(
        redshift=0.0, inputs=inputs, buffer_size=1
    )
    assert isinstance(pert_halo_cat.fesc_sfr, Array)
    inputs = inputs.evolve_input_structs(USE_MINI_HALOS=True)
    pert_halo_cat = ox.PerturbedHaloCatalog.new(
        redshift=0.0, inputs=inputs, buffer_size=1
    )
    assert isinstance(pert_halo_cat.stellar_mini, Array)
    assert isinstance(pert_halo_cat.sfr_mini, Array)


def test_optional_field_halobox(default_input_struct_lc: InputParameters):
    """Ensure that the correct HaloBox fields are set based on the parameters."""
    hb = ox.HaloBox.new(redshift=0.0, inputs=default_input_struct_lc)
    assert hb.halo_mass is None
    assert isinstance(hb.halo_sfr, Array)
    assert isinstance(hb.n_ion, Array)
    assert hb.halo_sfr_mini is None
    assert hb.halo_xray is None
    assert hb.whalo_sfr is None

    with config.use(EXTRA_HALOBOX_FIELDS=True):
        hb = ox.HaloBox.new(redshift=0.0, inputs=default_input_struct_lc)
        assert isinstance(hb.halo_mass, Array)
        assert isinstance(hb.count, Array)

        inputs = default_input_struct_lc.evolve_input_structs(INHOMO_RECO=True)
        hb = ox.HaloBox.new(redshift=0.0, inputs=inputs)
        assert isinstance(hb.whalo_sfr, Array)

        inputs = inputs.evolve_input_structs(USE_TS_FLUCT=True)
        hb = ox.HaloBox.new(redshift=0.0, inputs=inputs)
        assert isinstance(hb.halo_xray, Array)

        inputs = inputs.evolve_input_structs(USE_MINI_HALOS=True)
        hb = ox.HaloBox.new(redshift=0.0, inputs=inputs)
        assert isinstance(hb.halo_sfr_mini, Array)


def test_optional_field_xrs(default_input_struct_lc: InputParameters):
    """Ensure that the correct XraySourceBox fields are set based on the parameters."""
    xr = ox.XraySourceBox.new(redshift=0.0, inputs=default_input_struct_lc)
    assert isinstance(xr.filtered_sfr, Array)
    assert isinstance(xr.filtered_xray, Array)
    assert xr.filtered_sfr_mini is None

    inputs = default_input_struct_lc.evolve_input_structs(
        USE_TS_FLUCT=True,
        USE_MINI_HALOS=True,
        INHOMO_RECO=True,
    )
    xr = ox.XraySourceBox.new(redshift=0.0, inputs=inputs)
    assert isinstance(xr.filtered_sfr_mini, Array)


def test_optional_field_ts(default_input_struct_lc: InputParameters):
    """Ensure that the correct TsBox fields are set based on the parameters."""
    ts = ox.TsBox.new(redshift=0.0, inputs=default_input_struct_lc)
    assert isinstance(ts.spin_temperature, Array)
    assert isinstance(ts.xray_ionised_fraction, Array)
    assert isinstance(ts.kinetic_temp_neutral, Array)
    assert ts.J_21_LW is None

    inputs = default_input_struct_lc.evolve_input_structs(
        USE_TS_FLUCT=True,
        INHOMO_RECO=True,
        USE_MINI_HALOS=True,
    )
    ts = ox.TsBox.new(redshift=0.0, inputs=inputs)
    assert isinstance(ts.J_21_LW, Array)


def test_optional_field_ion(default_input_struct_lc: InputParameters):
    """Ensure that the correct IonizedBox fields are set based on the parameters."""
    ion = ox.IonizedBox.new(redshift=0.0, inputs=default_input_struct_lc)
    assert isinstance(ion.neutral_fraction, Array)
    assert ion.unnormalised_nion_mini is None
    assert ion.cumulative_recombinations is None

    inputs = default_input_struct_lc.evolve_input_structs(
        INHOMO_RECO=True,
    )
    ion = ox.IonizedBox.new(redshift=0.0, inputs=inputs)
    assert isinstance(ion.cumulative_recombinations, Array)

    inputs = inputs.evolve_input_structs(
        USE_TS_FLUCT=True,
        USE_MINI_HALOS=True,
    )
    ion = ox.IonizedBox.new(redshift=0.0, inputs=inputs)
    assert isinstance(ion.unnormalised_nion_mini, Array)


def test_optional_field_bt(default_input_struct_lc: InputParameters):
    """Ensure that the correct BrightnessTemp fields are set based on the parameters."""
    bt = ox.BrightnessTemp.new(redshift=0.0, inputs=default_input_struct_lc)
    assert isinstance(bt.brightness_temp, Array)
    assert bt.tau_21 is None

    inputs = default_input_struct_lc.evolve_input_structs(USE_TS_FLUCT=True)
    bt = ox.BrightnessTemp.new(redshift=0.0, inputs=inputs)
    assert isinstance(bt.tau_21, Array)


@pytest.mark.parametrize("struct", list(ox.OutputStructZ.__subclasses__()))
def test_bad_required_array(default_input_struct, struct):
    # no struct takes this input
    bt = ox.BrightnessTemp.new(redshift=10.0, inputs=default_input_struct)
    kwargs = {"inputs": default_input_struct, "redshift": 10.0}
    if struct is ox.PerturbedHaloCatalog:
        kwargs["buffer_size"] = 1
    output = struct.new(**kwargs)

    with pytest.raises(ValueError, match="is not an input required for"):
        _ = output.get_required_input_arrays(bt)



"""Test exceptions raised in the C backend."""

import numpy as np
import pytest
from py21cmfast.c_21cmfast import ffi, lib

from py21cmfast.wrapper.exceptions import (
    PHOTONCONSERROR,
    ParameterError,
    _process_exitcode,
)


@pytest.mark.parametrize("subfunc", [True, False])
def test_basic(subfunc):
    status = lib.SomethingThatCatches(subfunc)
    assert status == PHOTONCONSERROR


@pytest.mark.parametrize("subfunc", [True, False])
def test_simple(subfunc):
    answer = np.array([0], dtype="f8")

    status = lib.FunctionThatCatches(
        subfunc, False, ffi.cast("double *", ffi.from_buffer(answer))
    )

    with pytest.raises(ParameterError):
        _process_exitcode(
            status,
            lib.FunctionThatCatches,
            (False, ffi.cast("double *", ffi.from_buffer(answer))),
        )


def test_pass():
    answer = np.array([0], dtype="f8")
    lib.FunctionThatCatches(True, True, ffi.cast("double *", ffi.from_buffer(answer)))
    assert answer == 5.0



"""Test the Lightcone class."""

import numpy as np
import pytest
from astropy import units

from py21cmfast import InputParameters, LightCone
from py21cmfast.drivers.lightcone import _check_desired_arrays_exist


def mock_lightcone():
    inputs = InputParameters.from_template(
        "simple", node_redshifts=list(range(6, 40, 2)), random_seed=1
    ).evolve_input_structs(HII_DIM=200, BOX_LEN=200)
    rng = np.random.default_rng(123)

    dist = (
        np.arange(6000, 10000, inputs.simulation_options.cell_size.to_value("Mpc"))
        * units.Mpc
    )

    return LightCone(
        lightcone_distances=dist,
        inputs=inputs,
        lightcones={
            "brightness_temp": rng.uniform(
                -100,
                30,
                size=(
                    inputs.simulation_options.HII_DIM,
                    inputs.simulation_options.HII_DIM,
                    len(dist),
                ),
            )
        },
    )


class TestLightcone:
    """Test the LightCone object."""

    def test_invalid_quantity_raises(self):
        """Test that requesting an unknown lightcone quantity raises ValueError."""
        lc = mock_lightcone()
        with pytest.raises(ValueError, match="not computed"):
            _check_desired_arrays_exist(["not_a_real_quantity"], lc.inputs)

    def test_trim(self):
        """Test that trimming actually removes slices."""
        lc = mock_lightcone()

        newlc = lc.trim(7000 * units.Mpc, 9000 * units.Mpc)
        assert newlc.lightcone_distances.min() >= 7000 * units.Mpc
        assert newlc.lightcone_distances.max() <= 9000 * units.Mpc



"""Test configuration."""

import logging
import os
from pathlib import Path

import pytest

from py21cmfast import (
    AstroOptions,
    AstroParams,
    CosmoParams,
    InitialConditions,
    InputParameters,
    MatterOptions,
    OutputCache,
    PerturbedField,
    SimulationOptions,
    compute_initial_conditions,
    compute_ionization_field,
    config,
    get_logspaced_redshifts,
    perturb_field,
    run_lightcone,
)
from py21cmfast.io.caching import CacheConfig
from py21cmfast.lightconers import RectilinearLightconer


def pytest_addoption(parser):
    parser.addoption("--log-level-21", action="store", default="WARNING")


@pytest.fixture(scope="session")
def tmpdirec(tmp_path_factory):
    """Pytest fixture instantiating a new session-scope "data" folder.

    Parameters
    ----------
    tmpdir_factory :
        Pytest fixture for creating temporary directories.
    """
    return tmp_path_factory.mktemp("data")


def printdir(direc):
    try:
        width = os.get_terminal_size().columns
    except OSError:
        width = 100

    print()
    print(f" Files In {direc} ".center(width, "="))
    for pth in direc.iterdir():
        print(f"\t {pth.name:<20}:\t\t {pth.stat().st_size / 1024**2:.3f} KB")
    print("=" * width)


@pytest.fixture(scope="module")
def module_direc(tmp_path_factory):
    original = config["direc"]
    direc = tmp_path_factory.mktemp("modtmp")

    config["direc"] = str(direc)

    yield direc

    printdir(direc)

    # Set direc back to original.
    config["direc"] = original


@pytest.fixture
def test_direc(tmp_path_factory):
    original = config["direc"]
    direc = tmp_path_factory.mktemp("testtmp")

    config["direc"] = str(direc)

    yield direc

    printdir(direc)

    # Set direc back to original.
    config["direc"] = original


@pytest.fixture(autouse=True, scope="session")
def setup_and_teardown_package(tmpdirec, request):
    # Set nice global defaults for testing purposes, to make runs faster
    # (can always be over-ridden per-test).

    # Set default global parameters for all tests
    # ------ #

    # we run small boxes often here, and R_max is often large, so we ignore this error
    config["ignore_R_BUBBLE_MAX_error"] = True

    log_level = request.config.getoption("--log-level-21") or logging.INFO
    logging.getLogger("py21cmfast").setLevel(log_level)

    yield

    printdir(tmpdirec)


# ======================================================================================
# Create a default set of boxes that can be used throughout.
# ======================================================================================


@pytest.fixture(scope="session")
def default_seed():
    return 12


@pytest.fixture(scope="session")
def default_simulation_options():
    return SimulationOptions(
        HII_DIM=35,
        DIM=70,
        BOX_LEN=50,
        ZPRIME_STEP_FACTOR=1.2,
    )


@pytest.fixture(scope="session")
def default_matter_options():
    return MatterOptions(
        KEEP_3D_VELOCITIES=True,
        SOURCE_MODEL="E-INTEGRAL",  # We should consider which model is best for test defaults
    )


@pytest.fixture(scope="session")
def default_cosmo_params():
    return CosmoParams()


@pytest.fixture(scope="session")
def default_astro_options():
    return AstroOptions(
        USE_EXP_FILTER=False,
        CELL_RECOMB=False,
        USE_UPPER_STELLAR_TURNOVER=False,
    )


@pytest.fixture(scope="session")
def default_astro_options_ts():
    return AstroOptions(
        USE_EXP_FILTER=False,
        CELL_RECOMB=False,
        USE_TS_FLUCT=True,
        USE_UPPER_STELLAR_TURNOVER=False,
    )


@pytest.fixture(scope="session")
def default_astro_params():
    return AstroParams.new()


@pytest.fixture(scope="session")
def default_input_struct(
    default_simulation_options,
    default_matter_options,
    default_cosmo_params,
    default_astro_params,
    default_astro_options,
    default_seed,
):
    return InputParameters(
        random_seed=default_seed,
        cosmo_params=default_cosmo_params,
        astro_params=default_astro_params,
        simulation_options=default_simulation_options,
        matter_options=default_matter_options,
        astro_options=default_astro_options,
        node_redshifts=(),
    )


@pytest.fixture(scope="session")
def default_input_struct_ts(redshift, default_input_struct, default_astro_options_ts):
    return default_input_struct.clone(
        astro_options=default_astro_options_ts,
        node_redshifts=get_logspaced_redshifts(
            min_redshift=redshift,
            max_redshift=default_input_struct.simulation_options.Z_HEAT_MAX,
            z_step_factor=default_input_struct.simulation_options.ZPRIME_STEP_FACTOR,
        ),
    )


@pytest.fixture(scope="session")
def default_input_struct_lc(lightcone_min_redshift, default_input_struct):
    return default_input_struct.clone(
        node_redshifts=get_logspaced_redshifts(
            min_redshift=lightcone_min_redshift,
            max_redshift=default_input_struct.simulation_options.Z_HEAT_MAX,
            z_step_factor=default_input_struct.simulation_options.ZPRIME_STEP_FACTOR,
        )
    )


@pytest.fixture(scope="session")
def cache(tmpdirec: Path):
    return OutputCache(tmpdirec)


@pytest.fixture(scope="session")
def ic(default_input_struct, cache) -> InitialConditions:
    return compute_initial_conditions(
        inputs=default_input_struct,
        write=True,
        cache=cache,
    )


@pytest.fixture(scope="session")
def redshift():
    """A default redshift to evaluate at. Not too high, not too low."""
    return 15


@pytest.fixture(scope="session")
def lightcone_min_redshift(redshift):
    return redshift + 0.1


@pytest.fixture(scope="session")
def max_redshift():
    """A default redshift to evaluate at. Not too high, not too low."""
    return 25


@pytest.fixture(scope="session")
def low_redshift():
    """A default redshift to evaluate at. Not too high, not too low."""
    return 8


@pytest.fixture(scope="session")
def perturbed_field(ic, redshift, cache):
    """A default PerturbedField."""
    return perturb_field(
        redshift=redshift,
        initial_conditions=ic,
        write=True,
        cache=cache,
    )


@pytest.fixture(scope="session")
def perturbed_field_lc(
    ic: InitialConditions,
    default_input_struct_lc: InputParameters,
    redshift: float,
    cache,
):
    """A default PerturbedField for a lightcone (which requires node_redshifts)."""
    return perturb_field(
        redshift=redshift,
        inputs=default_input_struct_lc,
        initial_conditions=ic,
        write=True,
        cache=cache,
    )


@pytest.fixture(scope="session")
def ionize_box(
    ic: InitialConditions,
    perturbed_field: PerturbedField,
    cache: OutputCache,
):
    """A default ionize_box."""
    return compute_ionization_field(
        initial_conditions=ic,
        perturbed_field=perturbed_field,
        write=True,
        cache=cache,
    )


@pytest.fixture(scope="session")
def rectlcn(
    lightcone_min_redshift,
    max_redshift,
    default_simulation_options,
    default_cosmo_params,
) -> RectilinearLightconer:
    return RectilinearLightconer.between_redshifts(
        min_redshift=lightcone_min_redshift,
        max_redshift=max_redshift,
        resolution=default_simulation_options.cell_size,
        cosmo=default_cosmo_params.cosmo,
    )


@pytest.fixture(scope="session")
def lc(rectlcn, ic, cache, default_input_struct_lc):
    return run_lightcone(
        lightconer=rectlcn,
        initial_conditions=ic,
        inputs=default_input_struct_lc,
        write=CacheConfig(),
        cache=cache,
        include_dvdr_in_tau21=False,
    )


@pytest.fixture(scope="session")
def tiny_inputs():
    return (
        InputParameters.from_template(["simple", "tiny"], random_seed=1234)
        .evolve_input_structs(
            PHOTON_CONS_TYPE="z-photoncons",
            Z_HEAT_MAX=20,
        )
        .with_logspaced_redshifts(
            zmin=10,
            zmax=20,
        )
    )


@pytest.fixture(scope="session")
def tiny_ics(tiny_inputs):
    return compute_initial_conditions(inputs=tiny_inputs)



"""
Parameter Templates.

This file contains parameter default templates for use with sets of parameters which
are set to reproduce a certain published result or which we know work together.

These should make it easier to define a run, choosing the closest template to your
desired parameters and altering as few as possible.
"""

import datetime
import logging
from collections import defaultdict
from collections.abc import Sequence
from pathlib import Path
from typing import Any, Literal

import tomlkit

from .input_serialization import deserialize_inputs, prepare_inputs_for_serialization
from .wrapper.inputs import InputParameters

TEMPLATE_PATH = Path(__file__).parent / "templates/"
MANIFEST = TEMPLATE_PATH / "manifest.toml"

logger = logging.getLogger(__name__)

TOMLMode = Literal["full", "minimal"]


def list_templates() -> list[dict]:
    """Return a list of the available templates."""
    with MANIFEST.open("r") as fl:
        manifest = tomlkit.load(fl)
    return manifest["templates"]


def load_template_file(template_name: str | Path):
    """
    Handle the loading of template TOML files.

    First it checks for a file with the name given,
    then it checks for a native template with that name,
    throwing an error if neither are found.
    """
    if (fname := Path(template_name)).is_file():
        with fname.open("r") as fl:
            return tomlkit.load(fl)

    with MANIFEST.open("r") as fl:
        manifest = tomlkit.load(fl)

    for manf_entry in manifest["templates"]:
        if template_name.casefold() in [x.casefold() for x in manf_entry["aliases"]]:
            with (TEMPLATE_PATH / manf_entry["file"]).open("r") as fl:
                return tomlkit.load(fl)

    message = (
        f"Template {template_name} not found on-disk or in native template aliases\n"
        + "Available native templates are:\n"
    )
    for manf_entry in manifest["templates"]:
        message += f"{manf_entry['name']}: {manf_entry['aliases']}\n"
        message += f"     {manf_entry['description']}:\n\n"

    raise ValueError(message)


def create_params_from_template(
    template_name: str | Path | Sequence[str | Path], **kwargs
) -> dict[str, dict[str, Any]]:
    """
    Construct the required InputStruct instances for a run from a given template.

    Parameters
    ----------
    template_name: str,
        defines the name/alias of the native template (see templates/manifest.toml for a list)
        alternatively, is the path to a TOML file containing tables titled [CosmoParams],
        [SimulationOptions], [AstroParams] and [AstroOptions] with parameter settings

    Other Parameters
    ----------------
    Any other parameter passed is considered to be a name and
    value of a parameter in any of the `InputStruct` subclasses,
    and will be used to over-ride the template values.

    Returns
    -------
    input_dict : dict containing:
        cosmo_params : CosmoParams
            Instance containing cosmological parameters
        simulation_options : SimulationOptions
            Instance containing general run parameters
        simulation_options : MatterOptions
            Instance containing general run flags and enums
        astro_params : AstroParams
            Instance containing astrophysical parameters
        astro_options : AstroOptions
            Instance containing astrophysical flags and enums
    """
    if isinstance(template_name, str | Path):
        templates = [template_name]
    else:
        templates = template_name

    full_template = defaultdict(dict)
    for tmpl in templates:
        thist = load_template_file(tmpl)
        for k, v in thist.items():
            full_template[k] |= v
    return deserialize_inputs(full_template, **kwargs)


def write_template(
    inputs: InputParameters, template_file: Path | str, mode: TOMLMode = "full"
):
    """Write a set of input parameters to a template file.

    Parameters
    ----------
    inputs
        The inputs to write to file.
    template_file
        The path of the output.
    """
    inputs_dct = prepare_inputs_for_serialization(inputs, mode=mode)
    inputs_dct.pop("CosmoTables", None)

    template_file = Path(template_file)
    doc = tomlkit.document()
    doc.add(
        tomlkit.comment(
            "This file was generated by py21cmfast.run_templates.write_template"
        )
    )
    doc.add(tomlkit.comment(f"Created on: {datetime.datetime.now().isoformat()}"))
    doc.update(inputs_dct)

    with template_file.open("w") as fl:
        tomlkit.dump(doc, fl)



"""Test of the Lightconer subclasses."""

import re
from dataclasses import dataclass

import numpy as np
import pytest
from astropy import units as un
from astropy.cosmology import Planck18, z_at_value
from astropy_healpix import HEALPix
from scipy.spatial.transform import Rotation

from py21cmfast import AstroOptions, CosmoParams, InputParameters, SimulationOptions
from py21cmfast import lightconers as lcn


@pytest.fixture(scope="module")
def rect_lcner():
    return lcn.RectilinearLightconer.between_redshifts(
        min_redshift=6.0,
        max_redshift=7.0,
        resolution=2.0 * un.Mpc,  # Mpc
        quantities=("brightness_temp",),
    )


@pytest.fixture(scope="module")
def ang_lcner():
    hp = HEALPix(nside=4)
    lon, lat = hp.healpix_to_lonlat(np.arange(hp.npix))
    return lcn.AngularLightconer.between_redshifts(
        latitude=lat.radian,
        longitude=lon.radian,
        min_redshift=6.0,
        max_redshift=7.0,
        resolution=2 * un.Mpc,
        quantities=("brightness_temp",),
    )


@dataclass
class MockCoeval:
    """A mock coeval object for testing."""

    redshift: float
    brightness_temp: np.ndarray
    simulation_options: SimulationOptions
    cosmo_params: CosmoParams
    astro_options: AstroOptions


def get_uniform_coeval(redshift, fill=1.0, BOX_LEN=100, HII_DIM=50):
    up = SimulationOptions(BOX_LEN=BOX_LEN, HII_DIM=HII_DIM)

    return MockCoeval(
        redshift=redshift,
        brightness_temp=fill * np.ones((up.HII_DIM, up.HII_DIM, up.HII_DIM)),
        simulation_options=up,
        cosmo_params=CosmoParams(),
        astro_options=AstroOptions(),
    )


def test_equality(
    rect_lcner: lcn.RectilinearLightconer,
):
    assert rect_lcner == rect_lcner


@pytest.mark.parametrize("lc", ["rect_lcner", "ang_lcner"])
def test_uniform_coevals(request, lc):
    """Test that uniform boxes interpolate the way we want."""
    lc = request.getfixturevalue(lc)
    z6 = get_uniform_coeval(redshift=6.0, fill=0)
    z7 = get_uniform_coeval(redshift=7.0, fill=1.0)

    _q, _idx, out = next(lc.make_lightcone_slices(z6, z7))

    assert np.all(out >= 0)
    assert np.all(out <= 1)
    assert np.all(out[..., 0] == 0)


def test_incompatible_coevals(rect_lcner):
    z6 = get_uniform_coeval(redshift=6.0, fill=0)
    z7 = get_uniform_coeval(redshift=7.0, fill=1.0)

    # artificially change a cosmo param
    orig = z7.cosmo_params.SIGMA_8
    z7.cosmo_params = z7.cosmo_params.clone(SIGMA_8=2 * orig)

    with pytest.raises(ValueError, match="c1 and c2 must have the same cosmo"):
        next(rect_lcner.make_lightcone_slices(z6, z7))

    z7.cosmo_params = z7.cosmo_params.clone(SIGMA_8=orig)

    orig = z7.simulation_options.BOX_LEN
    z7.simulation_options = z7.simulation_options.clone(
        BOX_LEN=2 * z7.simulation_options.BOX_LEN
    )

    with pytest.raises(
        ValueError, match="c1 and c2 must have the same user parameters"
    ):
        next(rect_lcner.make_lightcone_slices(z6, z7))


def test_coeval_redshifts_outside_box(rect_lcner):
    z0 = rect_lcner.lc_redshifts[0] - 3
    z1 = rect_lcner.lc_redshifts[0] - 1

    z6 = get_uniform_coeval(redshift=z0, fill=0)
    z7 = get_uniform_coeval(redshift=z1, fill=1.0)

    q, idx, out = next(rect_lcner.make_lightcone_slices(z6, z7))
    assert q is None
    assert len(idx) == 0
    assert out is None


def test_bad_coeval_sizes_for_redshift_interp(rect_lcner):
    c1 = np.zeros((10, 10, 10))
    c2 = np.zeros((20, 20, 20))

    with pytest.raises(
        ValueError, match="coeval_a and coeval_b must have the same shape"
    ):
        rect_lcner.redshift_interpolation(3.0, c1, c2, 1, 4)


def test_bad_kind_for_redshift_interp(rect_lcner):
    c1 = np.zeros((10, 10, 10))
    c2 = np.zeros((10, 10, 10))

    with pytest.raises(ValueError, match="kind must be 'mean' or 'mean_max'"):
        rect_lcner.redshift_interpolation(3.0, c1, c2, 1, 4, kind="min")


def test_bad_instantiation():
    with pytest.raises(
        ValueError, match="Either lc_distances or lc_redshifts must be provided"
    ):
        lcn.RectilinearLightconer()

    with pytest.raises(ValueError, match="lc_distances must be non-negative"):
        lcn.RectilinearLightconer(lc_distances=np.linspace(-1, 1, 10) * un.Mpc)

    with pytest.raises(ValueError, match="lc_redshifts must be non-negative"):
        lcn.RectilinearLightconer(lc_redshifts=np.linspace(-1, 1, 10))


def test_equal_cdist_endpoint():
    hp = HEALPix(nside=4)
    lon, lat = hp.healpix_to_lonlat(np.arange(hp.npix))

    d0 = Planck18.comoving_distance(6.0)
    res = 2 * un.Mpc
    d1 = d0 + 20 * res
    zmax = z_at_value(Planck18.comoving_distance, d1)

    lc = lcn.AngularLightconer.between_redshifts(
        latitude=lat.radian,
        longitude=lon.radian,
        min_redshift=6.0,
        max_redshift=zmax,
        resolution=res,
    )
    assert np.isclose(lc.lc_distances.max(), d1, atol=d1 / 20)


def test_angular_lightconer_bad_instantiation():
    hp = HEALPix(nside=4)
    lon, lat = hp.healpix_to_lonlat(np.arange(hp.npix))

    with pytest.raises(ValueError, match="longitude must be 1-dimensional"):
        lcn.AngularLightconer(
            longitude=lon[None, :].radian,
            latitude=lat.radian,
            lc_redshifts=np.linspace(6, 7, 10),
        )

    with pytest.raises(
        ValueError, match=re.escape("longitude must be in the range [0, 2pi]")
    ):
        lcn.AngularLightconer(
            longitude=lon.radian + 2 * np.pi,
            latitude=lat.radian,
            lc_redshifts=np.linspace(6, 7, 10),
        )

    with pytest.raises(
        ValueError, match="longitude and latitude must have the same shape"
    ):
        lcn.AngularLightconer(
            longitude=lon.radian,
            latitude=lat[None, :].radian,
            lc_redshifts=np.linspace(6, 7, 10),
        )


def test_rotation_equality():
    hp = HEALPix(nside=4)
    lon, lat = hp.healpix_to_lonlat(np.arange(hp.npix))

    lc1 = lcn.AngularLightconer.between_redshifts(
        latitude=lat.radian,
        longitude=lon.radian,
        min_redshift=6.0,
        max_redshift=7.0,
        resolution=2 * un.Mpc,
    )

    lc2 = lcn.AngularLightconer.between_redshifts(
        latitude=lat.radian,
        longitude=lon.radian,
        min_redshift=6.0,
        max_redshift=7.0,
        resolution=2 * un.Mpc,
    )

    assert lc1 == lc2

    rot = Rotation.from_euler("z", np.pi / 2)
    lc3 = lcn.AngularLightconer.between_redshifts(
        latitude=lat.radian,
        longitude=lon.radian,
        min_redshift=6.0,
        max_redshift=7.0,
        resolution=2 * un.Mpc,
        rotation=rot,
    )

    assert lc1 != lc3


def test_validation_options_angular(ang_lcner):
    inputs = InputParameters(node_redshifts=np.array([5.0, 8.0]), random_seed=42)
    inputs.evolve_input_structs(KEEP_3D_VELOCITIES=False)
    with pytest.raises(
        ValueError,
        match="To account for RSDs or velocity corrections in an angular lightcone, you need to set",
    ):
        ang_lcner.validate_options(
            inputs=inputs, include_dvdr_in_tau21=True, apply_rsds=True
        )



"""Exceptions raised when running C code."""

import logging

logger = logging.getLogger(__name__)


class ParameterError(RuntimeError):
    """An exception representing a bad choice of parameters."""

    default_message = "21cmFAST does not support this combination of parameters."

    def __init__(self, msg=None):
        super().__init__(msg or self.default_message)


class FatalCError(Exception):
    """An exception representing something going wrong in C."""

    default_message = "21cmFAST is exiting."

    def __init__(self, msg=None):
        super().__init__(msg or self.default_message)


class FileIOError(FatalCError):
    """An exception when an error occurs with file I/O."""

    default_message = "Expected file could not be found! (check the LOG for more info)"


class GSLError(ParameterError):
    """An exception when a GSL routine encounters an error."""

    default_message = "A GSL routine has errored! (check the LOG for more info)"


class ArgumentValueError(FatalCError):
    """An exception when a function takes an unexpected input."""

    default_message = "An incorrect argument has been defined or passed! (check the LOG for more info)"


class PhotonConsError(ParameterError):
    """An exception when the photon non-conservation correction routine errors."""

    default_message = "An error has occured with the Photon non-conservation correction! (check the LOG for more info)"


class TableGenerationError(ParameterError):
    """An exception when an issue arises populating one of the interpolation tables."""

    default_message = """An error has occured when generating an interpolation table!
                This has likely occured due to the choice of input AstroParams (check the LOG for more info)"""


class TableEvaluationError(ParameterError):
    """An exception when an issue arises populating one of the interpolation tables."""

    default_message = """An error has occured when evaluating an interpolation table!
                This can sometimes occur due to small boxes (either small DIM/HII_DIM or BOX_LEN) (check the LOG for more info)"""


class InfinityorNaNError(ParameterError):
    """An exception when an infinity or NaN is encountered in a calculated quantity."""

    default_message = """Something has returned an infinity or a NaN! This could be due to an issue with an
                input parameter choice (check the LOG for more info)"""


class MassDepZetaError(ParameterError):
    """An exception when determining the bisection for stellar mass/escape fraction."""

    default_message = """There is an issue with the choice of parameters under MASS_DEPENDENT_ZETA. Could be an issue with
                any of the chosen F_STAR10, ALPHA_STAR, F_ESC10 or ALPHA_ESC."""


class MemoryAllocError(FatalCError):
    """An exception when unable to allocated memory."""

    default_message = """An error has occured while attempting to allocate memory! (check the LOG for more info)"""


SUCCESS = 0
IOERROR = 1
GSLERROR = 2
VALUEERROR = 3
PHOTONCONSERROR = 4
TABLEGENERATIONERROR = 5
TABLEEVALUATIONERROR = 6
INFINITYORNANERROR = 7
MASSDEPZETAERROR = 8
MEMORYALLOCERROR = 9


def _process_exitcode(exitcode, fnc, args):
    """Determine what happens for different values of the (integer) exit code from a C function."""
    if exitcode != SUCCESS:
        logger.error(f"In function: {fnc.__name__}.  Arguments: {args}")

        if exitcode:
            try:
                raise {
                    IOERROR: FileIOError,
                    GSLERROR: GSLError,
                    VALUEERROR: ArgumentValueError,
                    PHOTONCONSERROR: PhotonConsError,
                    TABLEGENERATIONERROR: TableGenerationError,
                    TABLEEVALUATIONERROR: TableEvaluationError,
                    INFINITYORNANERROR: InfinityorNaNError,
                    MASSDEPZETAERROR: MassDepZetaError,
                    MEMORYALLOCERROR: MemoryAllocError,
                }[exitcode]
            except KeyError as e:  # pragma: no cover
                raise FatalCError(
                    "Unknown error in C. Please report this error!"
                ) from e  # Unknown C code



"""Test that the package data exists in the correct places."""

import pytest

from py21cmfast import DATA_PATH


@pytest.fixture(scope="module")
def datafiles():
    return list(DATA_PATH.glob("*.dat"))


def test_exists(datafiles):
    for fl in datafiles:
        assert fl.exists()

    assert (DATA_PATH / "x_int_tables").exists()
    assert (DATA_PATH / "x_int_tables").is_dir()


def test_readable(datafiles):
    for fname in datafiles:
        with fname.open() as fl:
            assert len(fl.read()) > 0



"""Test the universal configuration module."""

import pytest
import yaml

import py21cmfast as p21
from py21cmfast._cfg import Config, ConfigurationError


@pytest.fixture(scope="module")
def cfgdir(tmp_path_factory):
    return tmp_path_factory.mktemp("config_test_dir")


def test_config_write(cfgdir):
    with p21.config.use(direc=str(cfgdir)):
        p21.config.write(cfgdir / "config.yml")

    with (cfgdir / "config.yml").open() as fl:
        new_config = yaml.load(fl, Loader=yaml.FullLoader)

    # Test adding new kind of string alias
    new_config["boxdir"] = new_config["direc"]
    del new_config["direc"]

    with (cfgdir / "config.yml").open("w") as fl:
        yaml.dump(new_config, fl)

    with pytest.raises(ConfigurationError):
        new_config = Config.load(cfgdir / "config.yml")
